@@ -1,4 +1,4 @@
-/*  Copyright (C) 2015 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2015-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -66,6 +66,7 @@ static int l_help(lua_State *L)
 		"verbose(true|false)\n    toggle verbose mode\n"
 		"option(opt[, new_val])\n    get/set server option\n"
 		"mode(strict|normal|permissive)\n    set resolver strictness level\n"
+		"reorder_RR([true|false])\n    set/get reordering of RRs within RRsets\n"
 		"resolve(name, type[, class, flags, callback])\n    resolve query, callback when it's finished\n"
 		"todname(name)\n    convert name to wire format\n"
 		"tojson(val)\n    convert value to JSON\n"
@@ -129,26 +130,6 @@ static int l_setuser(lua_State *L)
 	return 1;
 }
 
-/** Return platform-specific versioned library name. */
-static int l_libpath(lua_State *L)
-{
-	int n = lua_gettop(L);
-	if (n < 2)
-		return 0;
-	auto_free char *lib_path = NULL;
-	const char *lib_name = lua_tostring(L, 1);
-	const char *lib_version = lua_tostring(L, 2);
-#if defined(__APPLE__)
-	lib_path = afmt("%s.%s.dylib", lib_name, lib_version);
-#elif _WIN32
-	lib_path = afmt("%s.dll", lib_name); /* Versioned in RC files */
-#else
-	lib_path = afmt("%s.so.%s", lib_name, lib_version);
-#endif
-	lua_pushstring(L, lib_path);
-	return 1;
-}
-
 /** Quit current executable. */
 static int l_quit(lua_State *L)
 {
@@ -160,18 +141,63 @@ static int l_quit(lua_State *L)
 static int l_verbose(lua_State *L)
 {
 	if (lua_isboolean(L, 1) || lua_isnumber(L, 1)) {
-		kr_debug_set(lua_toboolean(L, 1));
+		kr_verbose_set(lua_toboolean(L, 1));
 	}
-	lua_pushboolean(L, kr_debug_status());
+	lua_pushboolean(L, kr_verbose_status);
 	return 1;
 }
 
+char *engine_get_hostname(struct engine *engine) {
+	static char hostname_str[KNOT_DNAME_MAXLEN];
+	if (!engine) {
+		return NULL;
+	}
+
+	if (!engine->hostname) {
+		if (gethostname(hostname_str, sizeof(hostname_str)) != 0)
+			return NULL;
+		return hostname_str;
+	}
+	return engine->hostname;
+}
+
+int engine_set_hostname(struct engine *engine, const char *hostname) {
+	if (!engine || !hostname) {
+		return kr_error(EINVAL);
+	}
+
+	char *new_hostname = strdup(hostname);
+	if (!new_hostname) {
+		return kr_error(ENOMEM);
+	}
+	if (engine->hostname) {
+		free(engine->hostname);
+	}
+	engine->hostname = new_hostname;
+	network_new_hostname(&engine->net, engine);
+
+	return 0;
+}
+
 /** Return hostname. */
 static int l_hostname(lua_State *L)
 {
-	char host_str[KNOT_DNAME_MAXLEN];
-	gethostname(host_str, sizeof(host_str));
-	lua_pushstring(L, host_str);
+	struct engine *engine = engine_luaget(L);
+	if (lua_gettop(L) == 0) {
+		lua_pushstring(L, engine_get_hostname(engine));
+		return 1;
+	}
+	if ((lua_gettop(L) != 1) || !lua_isstring(L, 1)) {
+		lua_pushstring(L, "hostname takes at most one parameter: (\"fqdn\")");
+		lua_error(L);
+	}
+
+	if (engine_set_hostname(engine, lua_tostring(L, 1)) != 0) {
+		lua_pushstring(L, "setting hostname failed");
+		lua_error(L);
+	}
+
+	lua_pushstring(L, engine_get_hostname(engine));	
 	return 1;
 }
 
@@ -369,6 +395,7 @@ static int l_map(lua_State *L)
 		/* Read response */
 		uint32_t rlen = 0;
 		if (read(fd, &rlen, sizeof(rlen)) == sizeof(rlen)) {
+			expr_checked(rlen < UINT32_MAX);
 			auto_free char *rbuf = malloc(rlen + 1);
 			expr_checked(rbuf != NULL);
 			expr_checked(read(fd, rbuf, rlen) == rlen);
@@ -457,22 +484,15 @@ static int init_resolver(struct engine *engine)
 		return kr_error(ENOMEM);
 	}
 	knot_edns_init(engine->resolver.opt_rr, KR_EDNS_PAYLOAD, 0, KR_EDNS_VERSION, engine->pool);
+	/* Set default TLS padding */
+	engine->resolver.tls_padding = KR_DEFAULT_TLS_PADDING;
 	/* Set default root hints */
 	kr_zonecut_init(&engine->resolver.root_hints, (const uint8_t *)"", engine->pool);
 	kr_zonecut_set_sbelt(&engine->resolver, &engine->resolver.root_hints);
 	/* Open NS rtt + reputation cache */
-	engine->resolver.cache_rtt = mm_alloc(engine->pool, lru_size(kr_nsrep_lru_t, LRU_RTT_SIZE));
-	if (engine->resolver.cache_rtt) {
-		lru_init(engine->resolver.cache_rtt, LRU_RTT_SIZE);
-	}
-	engine->resolver.cache_rep = mm_alloc(engine->pool, lru_size(kr_nsrep_lru_t, LRU_REP_SIZE));
-	if (engine->resolver.cache_rep) {
-		lru_init(engine->resolver.cache_rep, LRU_REP_SIZE);
-	}
-	engine->resolver.cache_cookie = mm_alloc(engine->pool, lru_size(kr_cookie_lru_t, LRU_COOKIES_SIZE));
-	if (engine->resolver.cache_cookie) {
-		lru_init(engine->resolver.cache_cookie, LRU_COOKIES_SIZE);
-	}
+	lru_create(&engine->resolver.cache_rtt, LRU_RTT_SIZE, engine->pool, NULL);
+	lru_create(&engine->resolver.cache_rep, LRU_REP_SIZE, engine->pool, NULL);
+	lru_create(&engine->resolver.cache_cookie, LRU_COOKIES_SIZE, engine->pool, NULL);
 
 	/* Load basic modules */
 	engine_register(engine, "iterate", NULL, NULL);
@@ -508,8 +528,10 @@ static int init_state(struct engine *engine)
 	lua_setglobal(engine->L, "user");
 	lua_pushcfunction(engine->L, l_trustanchor);
 	lua_setglobal(engine->L, "trustanchor");
-	lua_pushcfunction(engine->L, l_libpath);
-	lua_setglobal(engine->L, "libpath");
+	lua_pushliteral(engine->L, libknot_SONAME);
+	lua_setglobal(engine->L, "libknot_SONAME");
+	lua_pushliteral(engine->L, libzscanner_SONAME);
+	lua_setglobal(engine->L, "libzscanner_SONAME");
 	lua_pushcfunction(engine->L, l_tojson);
 	lua_setglobal(engine->L, "tojson");
 	lua_pushcfunction(engine->L, l_map);
@@ -523,20 +545,17 @@ static int init_state(struct engine *engine)
 	return kr_ok();
 }
 
+static enum lru_apply_do update_stat_item(const char *key, uint len,
+						unsigned *rtt, void *baton)
+{
+	return *rtt > KR_NS_LONG ? LRU_APPLY_DO_EVICT : LRU_APPLY_DO_NOTHING;
+}
+/** @internal Walk RTT table, clearing all entries with bad score
+ *    to compensate for intermittent network issues or temporary bad behaviour. */
 static void update_state(uv_timer_t *handle)
 {
 	struct engine *engine = handle->data;
-
-	/* Walk RTT table, clearing all entries with bad score
-	 * to compensate for intermittent network issues or temporary bad behaviour. */
-	kr_nsrep_lru_t *table = engine->resolver.cache_rtt;
-	for (size_t i = 0; i < table->size; ++i) {
-		if (!table->slots[i].key)
-			continue;
-		if (table->slots[i].data > KR_NS_LONG) {
-			lru_evict(table, i);
-		}
-	}
+	lru_apply(engine->resolver.cache_rtt, update_stat_item, NULL);
 }
 
 int engine_init(struct engine *engine, knot_mm_t *pool)
@@ -589,9 +608,11 @@ void engine_deinit(struct engine *engine)
 	network_deinit(&engine->net);
 	kr_zonecut_deinit(&engine->resolver.root_hints);
 	kr_cache_close(&engine->resolver.cache);
-	lru_deinit(engine->resolver.cache_rtt);
-	lru_deinit(engine->resolver.cache_rep);
-	lru_deinit(engine->resolver.cache_cookie);
+
+	/* The lru keys are currently malloc-ated and need to be freed. */
+	lru_free(engine->resolver.cache_rtt);
+	lru_free(engine->resolver.cache_rep);
+	lru_free(engine->resolver.cache_cookie);
 
 	/* Clear IPC pipes */
 	for (size_t i = 0; i < engine->ipc_set.len; ++i) {
@@ -804,6 +825,9 @@ int engine_register(struct engine *engine, const char *name, const char *precede
 	/* Load Lua module if not a binary */
 	if (ret == kr_error(ENOENT)) {
 		ret = ffimodule_register_lua(engine, module, name);
+	} else if (ret == kr_error(ENOTSUP)) {
+		/* Print a more helpful message when module is linked against an old resolver ABI. */
+		fprintf(stderr, "[system] module '%s' links to unsupported ABI, please rebuild it\n", name);
 	}
 	if (ret != 0) {
 		free(module);
