@@ -1,4 +1,4 @@
-/*  Copyright (C) 2014 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -17,6 +17,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <getopt.h>
+#include <libgen.h>
 #include <uv.h>
 #include <assert.h>
 #include <contrib/cleanup.h>
@@ -47,6 +48,16 @@
 static bool g_quiet = false;
 static bool g_interactive = true;
 
+/* lua_pcall helper function */
+static inline char *lua_strerror(int lua_err) {
+	switch (lua_err) {
+	case LUA_ERRRUN: return "a runtime error";
+	case LUA_ERRMEM: return "memory allocation error.";
+	case LUA_ERRERR: return "error while running the error handler function.";
+	default: return "a unknown error";
+	}
+}
+
 /*
  * TTY control
  */
@@ -55,11 +66,19 @@ static void tty_read(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf)
 	/* Set output streams */
 	FILE *out = stdout, *outerr = stderr;
 	uv_os_fd_t stream_fd = 0;
-	uv_fileno((uv_handle_t *)stream, &stream_fd);
+	if (uv_fileno((uv_handle_t *)stream, &stream_fd)) {
+		uv_close((uv_handle_t *)stream, (uv_close_cb) free);
+		if (buf) {
+			free(buf->base);
+		}
+		return;
+	}
 	if (stream_fd != STDIN_FILENO) {
 		if (nread <= 0) { /* Close if disconnected */
 			uv_close((uv_handle_t *)stream, (uv_close_cb) free);
-			free(buf->base);
+			if (buf) {
+				free(buf->base);
+			}
 			return;
 		}
 		uv_os_fd_t dup_fd = dup(stream_fd);
@@ -100,7 +119,9 @@ static void tty_read(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf)
 		lua_settop(L, 0);
 	}
 	fflush(out);
-	free(buf->base);
+	if (buf) {
+		free(buf->base);
+	}
 	/* Close if redirected */
 	if (stream_fd != STDIN_FILENO) {
 		fclose(out); /* outerr is the same */
@@ -166,7 +187,12 @@ static void ipc_activity(uv_poll_t* handle, int status, int events)
 	/* Read expression from IPC pipe */
 	uint32_t len = 0;
 	if (ipc_readall(fd, (char *)&len, sizeof(len))) {
-		auto_free char *rbuf = malloc(len + 1);
+		auto_free char *rbuf = NULL;
+		if (len < UINT32_MAX) {
+			rbuf = malloc(len + 1);
+		} else {
+			errno = EINVAL;
+		}
 		if (!rbuf) {
 			kr_log_error("[system] ipc: %s\n", strerror(errno));
 			engine_stop(engine); /* Panic and stop this fork. */
@@ -259,7 +285,6 @@ static int fork_workers(fd_array_t *ipc_set, int forks)
 			array_clear(*ipc_set);
 			array_push(*ipc_set, sv[0]);
 			close(sv[1]);
-			kr_crypto_reinit();
 			return forks;
 		/* Parent process */
 		} else {
@@ -284,46 +309,17 @@ static void help(int argc, char *argv[])
 	       " -k, --keyfile=[path] File containing trust anchors (DS or DNSKEY).\n"
 	       " -f, --forks=N        Start N forks sharing the configuration.\n"
 	       " -q, --quiet          Quiet output, no prompt in interactive mode.\n"
-	       " -v, --verbose        Run in verbose mode.\n"
+	       " -v, --verbose        Run in verbose mode."
+#ifdef NOVERBOSELOG
+	           " (Recompile without -DNOVERBOSELOG to activate.)"
+#endif
+	           "\n"
 	       " -V, --version        Print version of the server.\n"
 	       " -h, --help           Print help and usage.\n"
 	       "Options:\n"
 	       " [rundir]             Path to the working directory (default: .)\n");
 }
 
-static struct worker_ctx *init_worker(struct engine *engine, knot_mm_t *pool, int worker_id, int worker_count)
-{
-	/* Load bindings */
-	engine_lualib(engine, "modules", lib_modules);
-	engine_lualib(engine, "net",     lib_net);
-	engine_lualib(engine, "cache",   lib_cache);
-	engine_lualib(engine, "event",   lib_event);
-	engine_lualib(engine, "worker",  lib_worker);
-
-	/* Create main worker. */
-	struct worker_ctx *worker = mm_alloc(pool, sizeof(*worker));
-	if(!worker) {
-		return NULL;
-	}
-	memset(worker, 0, sizeof(*worker));
-	worker->id = worker_id;
-	worker->count = worker_count;
-	worker->engine = engine;
-	worker_reserve(worker, MP_FREELIST_SIZE);
-	/* Register worker in Lua thread */
-	lua_pushlightuserdata(engine->L, worker);
-	lua_setglobal(engine->L, "__worker");
-	lua_getglobal(engine->L, "worker");
-	lua_pushnumber(engine->L, worker_id);
-	lua_setfield(engine->L, -2, "id");
-	lua_pushnumber(engine->L, getpid());
-	lua_setfield(engine->L, -2, "pid");
-	lua_pushnumber(engine->L, worker_count);
-	lua_setfield(engine->L, -2, "count");
-	lua_pop(engine->L, 1);
-	return worker;
-}
-
 static int run_worker(uv_loop_t *loop, struct engine *engine, fd_array_t *ipc_set, bool leader, int control_fd)
 {
 	/* Control sockets or TTY */
@@ -362,7 +358,7 @@ static int run_worker(uv_loop_t *loop, struct engine *engine, fd_array_t *ipc_se
 	}
 	memcpy(&engine->ipc_set, ipc_set, sizeof(*ipc_set));
 
-	tls_setup_logging(kr_debug_status());
+	tls_setup_logging(kr_verbose_status);
 	/* Notify supervisor. */
 #ifdef HAS_SYSTEMD
 	sd_notify(0, "READY=1");
@@ -396,7 +392,6 @@ int main(int argc, char **argv)
 	array_init(tls_fd_set);
 	char *keyfile = NULL;
 	const char *config = NULL;
-	char *keyfile_buf = NULL;
 	int control_fd = -1;
 
 	/* Long options. */
@@ -436,40 +431,20 @@ int main(int argc, char **argv)
 		case 'f':
 			g_interactive = false;
 			forks = atoi(optarg);
-			if (forks == 0) {
-				kr_log_error("[system] error '-f' requires number, not '%s'\n", optarg);
+			if (forks <= 0) {
+				kr_log_error("[system] error '-f' requires a positive"
+						" number, not '%s'\n", optarg);
 				return EXIT_FAILURE;
 			}
 			break;
 		case 'k':
-			keyfile_buf = malloc(PATH_MAX);
-			assert(keyfile_buf);
-			/* Check if the path is absolute */
-			if (optarg[0] == '/') {
-				keyfile = strdup(optarg);
-			} else {
-				/* Construct absolute path, the file may not exist */
-				keyfile = realpath(".", keyfile_buf);
-				if (keyfile) {
-					int len = strlen(keyfile);
-					int namelen = strlen(optarg);
-					if (len + namelen < PATH_MAX - 1) {
-						keyfile[len] = '/';
-						memcpy(keyfile + len + 1, optarg, namelen + 1);
-						keyfile = strdup(keyfile); /* Duplicate */
-					} else {
-						keyfile = NULL; /* Invalidate */
-					}
-				}
-			}
-			free(keyfile_buf);
-			if (!keyfile) {
-				kr_log_error("[system] keyfile '%s': not writeable\n", optarg);
-				return EXIT_FAILURE;
-			}
+			keyfile = optarg;
 			break;
 		case 'v':
-			kr_debug_set(true);
+			kr_verbose_set(true);
+#ifdef NOVERBOSELOG
+			kr_log_info("--verbose flag has no effect due to compilation with -DNOVERBOSELOG.\n");
+#endif
 			break;
 		case 'q':
 			g_quiet = true;
@@ -515,6 +490,7 @@ int main(int argc, char **argv)
 	/* Switch to rundir. */
 	if (optind < argc) {
 		const char *rundir = argv[optind];
+		/* FIXME: access isn't a good way if we start as root and drop privileges later */
 		if (access(rundir, W_OK) != 0) {
 			kr_log_error("[system] rundir '%s': %s\n", rundir, strerror(errno));
 			return EXIT_FAILURE;
@@ -541,8 +517,6 @@ int main(int argc, char **argv)
 	 }
 #endif
 
-	kr_crypto_init();
-
 	/* Connect forks with local socket */
 	fd_array_t ipc_set;
 	array_init(ipc_set);
@@ -552,6 +526,8 @@ int main(int argc, char **argv)
 		return EXIT_FAILURE;
 	}
 
+	kr_crypto_init();
+
 	/* Create a server engine. */
 	knot_mm_t pool = {
 		.ctx = mp_new (4096),
@@ -564,7 +540,7 @@ int main(int argc, char **argv)
 		return EXIT_FAILURE;
 	}
 	/* Create worker */
-	struct worker_ctx *worker = init_worker(&engine, &pool, fork_id, forks);
+	struct worker_ctx *worker = worker_create(&engine, &pool, fork_id, forks);
 	if (!worker) {
 		kr_log_error("[system] not enough memory\n");
 		return EXIT_FAILURE;
@@ -614,6 +590,10 @@ int main(int argc, char **argv)
 		}
 	}
 
+	if (ret != 0) {
+		goto cleanup;
+	}
+
 	/* Block signals. */
 	uv_loop_t *loop = uv_default_loop();
 	uv_signal_t sigint, sigterm;
@@ -624,27 +604,99 @@ int main(int argc, char **argv)
 	/* Start the scripting engine */
 	worker->loop = loop;
 	loop->data = worker;
-	if (ret == 0) {
-		ret = engine_start(&engine, config ? config : "config");
-		if (ret == 0) {
-			if (keyfile) {
-				auto_free char *cmd = afmt("trust_anchors.config('%s')", keyfile);
-				if (!cmd) {
-					kr_log_error("[system] not enough memory\n");
-					return EXIT_FAILURE;
-				}
-				engine_cmd(engine.L, cmd, false);
-				lua_settop(engine.L, 0);
-			}
-			/* Run the event loop */
-			ret = run_worker(loop, &engine, &ipc_set, fork_id == 0, control_fd);
-		}
+
+	ret = engine_start(&engine, config ? config : "config");
+	if (ret != 0) {
+		ret = EXIT_FAILURE;
+		goto cleanup;
 	}
+
+	if (keyfile) {
+		auto_free char *dirname_storage = strdup(keyfile);
+		if (!dirname_storage) {
+			kr_log_error("[system] not enough memory: %s\n",
+				     strerror(errno));
+			ret = EXIT_FAILURE;
+			goto cleanup;
+		}
+		auto_free char *basename_storage = strdup(keyfile);
+		if (!basename_storage) {
+			kr_log_error("[system] not enough memory: %s\n",
+				     strerror(errno));
+			ret = EXIT_FAILURE;
+			goto cleanup;
+		}
+
+		/* Resolve absolute path to the keyfile directory */
+		auto_free char *keyfile_dir = malloc(PATH_MAX);
+		if (realpath(dirname(dirname_storage), keyfile_dir) == NULL) {
+			kr_log_error("[ ta ]: keyfile '%s' directory: %s\n",
+				     keyfile, strerror(errno));
+			ret = EXIT_FAILURE;
+			goto cleanup;
+		}
+
+		char *_filename = basename(basename_storage);
+		int dirlen = strlen(keyfile_dir);
+		int namelen = strlen(_filename);
+		if (dirlen + namelen >= PATH_MAX) {
+			kr_log_error("[ ta ]: keyfile '%s' PATH_MAX exceeded\n",
+				     keyfile);
+			ret = EXIT_FAILURE;
+			goto cleanup;
+		}
+		keyfile_dir[dirlen] = '/';
+
+		auto_free char *keyfile_path = malloc(dirlen + namelen + 1);
+		memcpy(keyfile_path, keyfile_dir, dirlen + 1);
+		memcpy(keyfile_path + dirlen + 1, _filename, namelen + 1);
+
+		int unmanaged = 0;
+
+		/* Note: config has been executed, so access() is OK,
+		 * as we've dropped privileges already if configured. */
+		if (access(keyfile_path, F_OK) != 0) {
+			kr_log_info("[ ta ] keyfile '%s': doesn't exist, bootstrapping\n", keyfile_path);
+			if (access(keyfile_dir, W_OK) != 0) {
+				kr_log_error("[ ta ] keyfile '%s': write access to '%s' needed\n", keyfile_path, keyfile_dir);
+				ret = EXIT_FAILURE;
+				goto cleanup;
+			}
+		} else if (access(keyfile_path, R_OK) == 0) {
+			if ((access(keyfile_path, W_OK) != 0) || (access(keyfile_dir, W_OK) != 0)) {
+				kr_log_error("[ ta ] keyfile '%s': not writeable, starting in unmanaged mode\n", keyfile_path);
+				unmanaged = 1;
+			}
+		} else {
+			kr_log_error("[ ta ] keyfile '%s': %s\n", keyfile_path, strerror(errno));
+			ret = EXIT_FAILURE;
+			goto cleanup;
+		}
+
+		auto_free char *cmd = afmt("trust_anchors.config('%s',%s)", keyfile_path, unmanaged?"true":"nil");
+		if (!cmd) {
+			kr_log_error("[system] not enough memory\n");
+			ret =  EXIT_FAILURE;
+			goto cleanup;
+		}
+		int lua_ret = 0;
+		if ((lua_ret = engine_cmd(engine.L, cmd, false)) != 0) {
+			kr_log_error("[ ta ] keyfile '%s': failed to load (%s)\n", keyfile_path, lua_strerror(lua_ret));
+			ret = EXIT_FAILURE;
+			goto cleanup;
+		}
+		lua_settop(engine.L, 0);
+	}
+
+	/* Run the event loop */
+	ret = run_worker(loop, &engine, &ipc_set, fork_id == 0, control_fd);
 	if (ret != 0) {
 		perror("[system] worker failed");
 		ret = EXIT_FAILURE;
+		goto cleanup;
 	}
-	/* Cleanup. */
+
+cleanup:/* Cleanup. */
 	engine_deinit(&engine);
 	worker_reclaim(worker);
 	mp_delete(pool.ctx);
