@@ -1,4 +1,4 @@
-/*  Copyright (C) 2014 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -16,6 +16,7 @@
 
 #include <sys/time.h>
 #include <assert.h>
+#include <arpa/inet.h>
 
 #include <libknot/descriptor.h>
 #include <libknot/rrtype/rdname.h>
@@ -29,7 +30,8 @@
 #include "lib/module.h"
 #include "lib/dnssec/ta.h"
 
-#define DEBUG_MSG(fmt...) QRDEBUG(req->current_query, "iter", fmt)
+#define VERBOSE_MSG(fmt...) QRVERBOSE(req->current_query, "iter", fmt)
+#define QVERBOSE_MSG(qry, fmt...) QRVERBOSE(qry, "iter", fmt)
 
 /* Iterator often walks through packet section, this is an abstraction. */
 typedef int (*rr_callback_t)(const knot_rrset_t *, unsigned, struct kr_request *);
@@ -67,7 +69,7 @@ static bool is_paired_to_query(const knot_pkt_t *answer, struct kr_query *query)
 
 	return query->id      == knot_wire_get_id(answer->wire) &&
 	       knot_wire_get_qdcount(answer->wire) > 0 &&
-	       (query->sclass == KNOT_CLASS_ANY || query->sclass  == knot_pkt_qclass(answer)) &&
+	       query->sclass  == knot_pkt_qclass(answer) &&
 	       qtype          == knot_pkt_qtype(answer) &&
 	       knot_dname_is_equal(qname, knot_pkt_qname(answer));
 }
@@ -131,66 +133,51 @@ static bool is_valid_addr(const uint8_t *addr, size_t len)
 	return true;
 }
 
+/** @internal Update NS address from record \a rr.  Return _FAIL on error. */
 static int update_nsaddr(const knot_rrset_t *rr, struct kr_query *query)
 {
 	if (rr->type == KNOT_RRTYPE_A || rr->type == KNOT_RRTYPE_AAAA) {
 		const knot_rdata_t *rdata = rr->rrs.data;
 		if (!(query->flags & QUERY_ALLOW_LOCAL) &&
 			!is_valid_addr(knot_rdata_data(rdata), knot_rdata_rdlen(rdata))) {
-			return KNOT_STATE_CONSUME; /* Ignore invalid addresses */
+			return KR_STATE_CONSUME; /* Ignore invalid addresses */
 		}
 		int ret = kr_zonecut_add(&query->zone_cut, rr->owner, rdata);
 		if (ret != 0) {
-			return KNOT_STATE_FAIL;
+			return KR_STATE_FAIL;
 		}
 	}
 
-	return KNOT_STATE_CONSUME;
-}
-
-static int update_parent(const knot_rrset_t *rr, struct kr_query *qry)
-{
-	return update_nsaddr(rr, qry->parent);
-}
-
-static int update_answer(const knot_rrset_t *rr, unsigned hint, knot_pkt_t *answer)
-{
-	/* Scrub DNSSEC records when not requested. */
-	if (!knot_pkt_has_dnssec(answer)) {
-		if (rr->type != knot_pkt_qtype(answer) && knot_rrtype_is_dnssec(rr->type)) {
-			return KNOT_STATE_DONE; /* Scrub */
-		}
-	}
-	/* Copy record, as it may be accessed after packet processing. */
-	knot_rrset_t *copy = knot_rrset_copy(rr, &answer->mm);
-	/* Write to final answer. */
-	int ret = knot_pkt_put(answer, hint, copy, KNOT_PF_FREE);
-	if (ret != KNOT_EOK) {
-		knot_wire_set_tc(answer->wire);
-		return KNOT_STATE_DONE;
-	}
-
-	return KNOT_STATE_DONE;
+	return KR_STATE_CONSUME;
 }
 
+/** @internal From \a pkt, fetch glue records for name \a ns, and update the cut etc. */
 static void fetch_glue(knot_pkt_t *pkt, const knot_dname_t *ns, struct kr_request *req)
 {
-	bool used_glue = false;
 	for (knot_section_t i = KNOT_ANSWER; i <= KNOT_ADDITIONAL; ++i) {
 		const knot_pktsection_t *sec = knot_pkt_section(pkt, i);
 		for (unsigned k = 0; k < sec->count; ++k) {
 			const knot_rrset_t *rr = knot_pkt_rr(sec, k);
-			if (knot_dname_is_equal(ns, rr->owner)) {
-				(void) update_nsaddr(rr, req->current_query);
-				used_glue = true;
+			if (!knot_dname_is_equal(ns, rr->owner)) {
+				continue;
+			}
+			if ((rr->type != KNOT_RRTYPE_A) &&
+			    (rr->type != KNOT_RRTYPE_AAAA)) {
+				continue;
+			}
+			(void) update_nsaddr(rr, req->current_query);
+			WITH_VERBOSE {
+				char name_str[KNOT_DNAME_MAXLEN];
+				char addr_str[INET6_ADDRSTRLEN];
+				const void *addr = knot_rdata_data(rr->rrs.data);
+				const int addr_len = knot_rdata_rdlen(rr->rrs.data);
+				const int af = (addr_len == sizeof(struct in_addr)) ?
+					       AF_INET : AF_INET6;
+				knot_dname_to_str(name_str, ns, sizeof(name_str));
+				inet_ntop(af, addr, addr_str, sizeof(addr_str));
+				VERBOSE_MSG("<= using glue for '%s': '%s'\n",
+					    name_str, addr_str);
 			}
-		}
-	}
-	WITH_DEBUG {
-		char name_str[KNOT_DNAME_MAXLEN];
-		knot_dname_to_str(name_str, ns, sizeof(name_str));
-		if (used_glue) {
-			DEBUG_MSG("<= using glue for '%s'\n", name_str);
 		}
 	}
 }
@@ -211,30 +198,32 @@ static int has_glue(knot_pkt_t *pkt, const knot_dname_t *ns)
 	return 0;
 }
 
-static int update_cut(knot_pkt_t *pkt, const knot_rrset_t *rr, struct kr_request *req)
+/** @internal Update the cut with another NS(+glue) record.
+ * @param current_cut is cut name before this packet.
+ * @return _DONE if cut->name changes, _FAIL on error, and _CONSUME otherwise. */
+static int update_cut(knot_pkt_t *pkt, const knot_rrset_t *rr,
+		      struct kr_request *req, const knot_dname_t *current_cut)
 {
 	struct kr_query *qry = req->current_query;
 	struct kr_zonecut *cut = &qry->zone_cut;
-	int state = KNOT_STATE_CONSUME;
+	int state = KR_STATE_CONSUME;
 
-	/* Authority MUST be at/below the authority of the nameserver, otherwise
-	 * possible cache injection attempt. */
-	if (!knot_dname_in(cut->name, rr->owner)) {
-		DEBUG_MSG("<= authority: ns outside bailiwick\n");
+	/* New authority MUST be at/below the authority of the current cut;
+	 * otherwise it's a possible cache injection attempt. */
+	if (!knot_dname_in(current_cut, rr->owner)) {
+		VERBOSE_MSG("<= authority: ns outside bailiwick\n");
 #ifdef STRICT_MODE
-		return KNOT_STATE_FAIL;
+		return KR_STATE_FAIL;
 #else
 		/* Workaround: ignore out-of-bailiwick NSs for authoritative answers,
 		 * but fail for referrals. This is important to detect lame answers. */
 		if (knot_pkt_section(pkt, KNOT_ANSWER)->count == 0) {
-			state = KNOT_STATE_FAIL;
+			state = KR_STATE_FAIL;
 		}
 		return state;
 #endif
 	}
 
-	/* Remember current bailiwick for NS processing. */
-	const knot_dname_t *current_cut = cut->name;
 	/* Update zone cut name */
 	if (!knot_dname_is_equal(rr->owner, cut->name)) {
 		/* Remember parent cut and descend to new (keep keys and TA). */
@@ -248,7 +237,7 @@ static int update_cut(knot_pkt_t *pkt, const knot_rrset_t *rr, struct kr_request
 		} else {
 			kr_zonecut_set(cut, rr->owner);
 		}
-		state = KNOT_STATE_DONE;
+		state = KR_STATE_DONE;
 	}
 
 	/* Fetch glue for each NS */
@@ -257,7 +246,7 @@ static int update_cut(knot_pkt_t *pkt, const knot_rrset_t *rr, struct kr_request
 		int glue_records = has_glue(pkt, ns_name);
 		/* Glue is mandatory for NS below zone */
 		if (!glue_records && knot_dname_in(rr->owner, ns_name)) {
-			DEBUG_MSG("<= authority: missing mandatory glue, rejecting\n");
+			VERBOSE_MSG("<= authority: missing mandatory glue, rejecting\n");
 			continue;
 		}
 		kr_zonecut_add(cut, ns_name, NULL);
@@ -278,41 +267,70 @@ static int update_cut(knot_pkt_t *pkt, const knot_rrset_t *rr, struct kr_request
 	return state;
 }
 
-static int process_authority(knot_pkt_t *pkt, struct kr_request *req)
+static int pick_authority(knot_pkt_t *pkt, struct kr_request *req, bool to_wire)
 {
-	int result = KNOT_STATE_CONSUME;
 	struct kr_query *qry = req->current_query;
 	const knot_pktsection_t *ns = knot_pkt_section(pkt, KNOT_AUTHORITY);
-
-	/* Stub resolution doesn't process authority */
-	if (qry->flags & QUERY_STUB) {
-		return KNOT_STATE_CONSUME;
+	uint8_t rank = !(qry->flags & QUERY_DNSSEC_WANT) || (qry->flags & QUERY_CACHED) ?
+			KR_VLDRANK_SECURE : KR_VLDRANK_INITIAL;
+	const knot_dname_t *zonecut_name = qry->zone_cut.name;
+	bool referral = !knot_wire_get_aa(pkt->wire);
+	if (referral) {
+		/* zone cut already updated by process_authority()
+		 * use parent zonecut name */
+		zonecut_name = qry->zone_cut.parent ? qry->zone_cut.parent->name : qry->zone_cut.name;
+		to_wire = false;
 	}
 
+	for (unsigned i = 0; i < ns->count; ++i) {
+		const knot_rrset_t *rr = knot_pkt_rr(ns, i);
+		if (!knot_dname_in(zonecut_name, rr->owner)) {
+			continue;
+		}
+		int ret = kr_ranked_rrarray_add(&req->auth_selected, rr,
+						rank, to_wire, qry->uid, &req->pool);
+		if (ret != kr_ok()) {
+			return ret;
+		}
+	}
+
+	return kr_ok();
+}
+
+static int process_authority(knot_pkt_t *pkt, struct kr_request *req)
+{
+	struct kr_query *qry = req->current_query;
+	assert(!(qry->flags & QUERY_STUB));
+
+	int result = KR_STATE_CONSUME;
+	const knot_pktsection_t *ns = knot_pkt_section(pkt, KNOT_AUTHORITY);
+
 #ifdef STRICT_MODE
 	/* AA, terminate resolution chain. */
 	if (knot_wire_get_aa(pkt->wire)) {
-		return KNOT_STATE_CONSUME;
+		return KR_STATE_CONSUME;
 	}
 #else
+
 	/* Work around servers sending back CNAME with different delegation and no AA. */
 	const knot_pktsection_t *an = knot_pkt_section(pkt, KNOT_ANSWER);
 	if (an->count > 0 && ns->count > 0) {
 		const knot_rrset_t *rr = knot_pkt_rr(an, 0);
 		if (rr->type == KNOT_RRTYPE_CNAME) {
-			return KNOT_STATE_CONSUME;
+			return KR_STATE_CONSUME;
 		}
 	}
 #endif
-
+	/* Remember current bailiwick for NS processing. */
+	const knot_dname_t *current_zone_cut = qry->zone_cut.name;
 	/* Update zone cut information. */
 	for (unsigned i = 0; i < ns->count; ++i) {
 		const knot_rrset_t *rr = knot_pkt_rr(ns, i);
 		if (rr->type == KNOT_RRTYPE_NS) {
-			int state = update_cut(pkt, rr, req);
+			int state = update_cut(pkt, rr, req, current_zone_cut);
 			switch(state) {
-			case KNOT_STATE_DONE: result = state; break;
-			case KNOT_STATE_FAIL: return state; break;
+			case KR_STATE_DONE: result = state; break;
+			case KR_STATE_FAIL: return state; break;
 			default:              /* continue */ break;
 			}
 		} else if (rr->type == KNOT_RRTYPE_SOA && knot_dname_is_sub(rr->owner, qry->zone_cut.name)) {
@@ -321,6 +339,13 @@ static int process_authority(knot_pkt_t *pkt, struct kr_request *req)
 		}
 	}
 
+	if ((qry->flags & QUERY_DNSSEC_WANT) && (result == KR_STATE_CONSUME)) {
+		if (knot_wire_get_aa(pkt->wire) == 0 && knot_wire_get_ancount(pkt->wire) == 0) {
+			/* Prevent from validating as an autoritative answer */
+			result = KR_STATE_DONE;
+		}
+	}
+
 	/* CONSUME => Unhelpful referral.
 	 * DONE    => Zone cut updated.  */
 	return result;
@@ -331,26 +356,6 @@ static void finalize_answer(knot_pkt_t *pkt, struct kr_query *qry, struct kr_req
 	/* Finalize header */
 	knot_pkt_t *answer = req->answer;
 	knot_wire_set_rcode(answer->wire, knot_wire_get_rcode(pkt->wire));
-
-	/* Fill in bailiwick records in authority */
-	const bool scrub_dnssec = !knot_pkt_has_dnssec(answer);
-	const uint16_t qtype = knot_pkt_qtype(answer);
-	struct kr_zonecut *cut = &qry->zone_cut;
-	int pkt_class = kr_response_classify(pkt);
-	if ((pkt_class & (PKT_NXDOMAIN|PKT_NODATA))) {
-		const knot_pktsection_t *ns = knot_pkt_section(pkt, KNOT_AUTHORITY);
-		for (unsigned i = 0; i < ns->count; ++i) {
-			const knot_rrset_t *rr = knot_pkt_rr(ns, i);
-			/* Scrub DNSSEC records when not requested. */
-			if (scrub_dnssec && rr->type != qtype && knot_rrtype_is_dnssec(rr->type)) {
-				continue;
-			}
-			/* Stash the authority records, they will be written to wire on answer finalization. */
-			if (knot_dname_in(cut->name, rr->owner)) {
-				kr_rrarray_add(&req->authority, rr, &answer->mm);
-			}
-		}
-	}
 }
 
 static bool is_rrsig_type_covered(const knot_rrset_t *rr, uint16_t type)
@@ -365,36 +370,18 @@ static bool is_rrsig_type_covered(const knot_rrset_t *rr, uint16_t type)
 	return ret;
 }
 
-static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
+static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral, const knot_dname_t **cname_ret)
 {
 	struct kr_query *query = req->current_query;
-	/* Response for minimized QNAME.
-	 * NODATA   => may be empty non-terminal, retry (found zone cut)
-	 * NOERROR  => found zone cut, retry
-	 * NXDOMAIN => parent is zone cut, retry as a workaround for bad authoritatives
-	 */
-	bool is_final = (query->parent == NULL);
-	int pkt_class = kr_response_classify(pkt);
-	if (!knot_dname_is_equal(knot_pkt_qname(pkt), query->sname) &&
-	    (pkt_class & (PKT_NOERROR|PKT_NXDOMAIN|PKT_REFUSED|PKT_NODATA))) {
-		DEBUG_MSG("<= found cut, retrying with non-minimized name\n");
-		query->flags |= QUERY_NO_MINIMIZE;
-		return KNOT_STATE_CONSUME;
-	}
-
-	/* This answer didn't improve resolution chain, therefore must be authoritative (relaxed to negative). */
-	if (!(query->flags & QUERY_STUB) && !is_authoritative(pkt, query)) {
-		if (pkt_class & (PKT_NXDOMAIN|PKT_NODATA)) {
-			DEBUG_MSG("<= lame response: non-auth sent negative response\n");
-			return KNOT_STATE_FAIL;
-		}
-	}
-
+	assert(!(query->flags & QUERY_STUB));
 	/* Process answer type */
 	const knot_pktsection_t *an = knot_pkt_section(pkt, KNOT_ANSWER);
 	const knot_dname_t *cname = NULL;
 	const knot_dname_t *pending_cname = query->sname;
 	unsigned cname_chain_len = 0;
+	uint8_t rank = !(query->flags & QUERY_DNSSEC_WANT) || (query->flags & QUERY_CACHED) ?
+			KR_VLDRANK_SECURE : KR_VLDRANK_INITIAL;
+	bool is_final = (query->parent == NULL);
 	bool can_follow = false;
 	bool strict_mode = (query->flags & QUERY_STRICT);
 	do {
@@ -410,13 +397,21 @@ static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
 				continue;
 			}
 			/* Process records matching current SNAME */
-			unsigned hint = 0;
-			if(knot_dname_is_equal(cname, knot_pkt_qname(req->answer))) {
-				hint = KNOT_COMPR_HINT_QNAME;
+			int state = KR_STATE_FAIL;
+			bool to_wire = false;
+			if (is_final) {
+				/* if not referral, mark record to be written to final answer */
+				to_wire = !referral;
+			} else {
+				state = update_nsaddr(rr, query->parent);
+				if (state == KR_STATE_FAIL) {
+					return state;
+				}
 			}
-			int state = is_final ? update_answer(rr, hint, req->answer) : update_parent(rr, query);
-			if (state == KNOT_STATE_FAIL) {
-				return state;
+			state = kr_ranked_rrarray_add(&req->answ_selected, rr,
+						      rank, to_wire, query->uid, &req->pool);
+			if (state != kr_ok()) {
+				return KR_STATE_FAIL;
 			}
 			/* can_follow is false, therefore QUERY_DNSSEC_WANT flag is set.
 			 * Follow cname chain only if rrsig exists. */
@@ -430,23 +425,126 @@ static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
 			}
 			cname_chain_len += 1;
 			pending_cname = knot_cname_name(&rr->rrs);
-			if (!pending_cname || strict_mode) {
+			if (!pending_cname) {
 				break;
 			}
 			if (cname_chain_len > an->count || cname_chain_len > KR_CNAME_CHAIN_LIMIT) {
-				DEBUG_MSG("<= too long cname chain\n");
-				return KNOT_STATE_FAIL;
+				VERBOSE_MSG("<= too long cname chain\n");
+				return KR_STATE_FAIL;
 			}
 			/* Don't use pending_cname immediately.
 			 * There are can be records for "old" cname. */
 		}
+		if (!pending_cname) {
+			break;
+		}
+		if (knot_dname_is_equal(cname, pending_cname)) {
+			VERBOSE_MSG("<= cname chain loop\n");
+			return KR_STATE_FAIL;
+		}
 		/* In strict mode, explicitly fetch each CNAME target. */
-		if (strict_mode && pending_cname) {
+		if (strict_mode) {
 			cname = pending_cname;
 			break;
 		}
-	} while (pending_cname && can_follow);
+		/* try to unroll cname only within current zone */
+		const int pending_labels = knot_dname_labels(pending_cname, NULL);
+		const int cname_labels = knot_dname_labels(cname, NULL);
+		if (pending_labels != cname_labels) {
+			cname = pending_cname;
+			break;
+		}
+		if (knot_dname_matched_labels(pending_cname, cname) !=
+		    (cname_labels - 1)) {
+			cname = pending_cname;
+			break;
+		}
+	} while (can_follow);
+	*cname_ret = cname;
+	return kr_ok();
+}
 
+static int process_referral_answer(knot_pkt_t *pkt, struct kr_request *req)
+{
+	const knot_dname_t *cname = NULL;
+	int state = unroll_cname(pkt, req, true, &cname);
+	if (state != kr_ok()) {
+		return KR_STATE_FAIL;
+	}
+	state = pick_authority(pkt, req, false);
+	return state == kr_ok() ? KR_STATE_DONE : KR_STATE_FAIL;
+}
+
+static int process_final(knot_pkt_t *pkt, struct kr_request *req,
+			 const knot_dname_t *cname)
+{
+	const int pkt_class = kr_response_classify(pkt);
+	struct kr_query *query = req->current_query;
+	ranked_rr_array_t *array = &req->answ_selected;
+	for (size_t i = 0; i < array->len; ++i) {
+		const knot_rrset_t *rr = array->at[i]->rr;
+		if (!knot_dname_is_equal(rr->owner, cname)) {
+			continue;
+		}
+		if ((rr->rclass != query->sclass) ||
+		    (rr->type != query->stype)) {
+			continue;
+		}
+		const bool to_wire = ((pkt_class & (PKT_NXDOMAIN|PKT_NODATA)) != 0);
+		const int state = pick_authority(pkt, req, to_wire);
+		if (state != kr_ok()) {
+			return KR_STATE_FAIL;
+		}
+		if (!array->at[i]->to_wire) {
+			const size_t last_idx = array->len - 1;
+			size_t j = i;
+			ranked_rr_array_entry_t *entry = array->at[i];
+			/* Relocate record to the end, after current cname */
+			while (j < last_idx) {
+				array->at[j] = array->at[j + 1];
+				++j;
+			}
+			array->at[last_idx] = entry;
+			entry->to_wire = true;
+		}
+		finalize_answer(pkt, query, req);
+		return KR_STATE_DONE;
+	}
+	return kr_ok();
+}
+
+static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
+{
+	struct kr_query *query = req->current_query;
+
+	/* Response for minimized QNAME.
+	 * NODATA   => may be empty non-terminal, retry (found zone cut)
+	 * NOERROR  => found zone cut, retry
+	 * NXDOMAIN => parent is zone cut, retry as a workaround for bad authoritatives
+	 */
+	bool is_final = (query->parent == NULL);
+	int pkt_class = kr_response_classify(pkt);
+	if (!knot_dname_is_equal(knot_pkt_qname(pkt), query->sname) &&
+	    (pkt_class & (PKT_NOERROR|PKT_NXDOMAIN|PKT_REFUSED|PKT_NODATA))) {
+		VERBOSE_MSG("<= found cut, retrying with non-minimized name\n");
+		query->flags |= QUERY_NO_MINIMIZE;
+		return KR_STATE_CONSUME;
+	}
+
+	/* This answer didn't improve resolution chain, therefore must be authoritative (relaxed to negative). */
+	if (!is_authoritative(pkt, query)) {
+		if (pkt_class & (PKT_NXDOMAIN|PKT_NODATA)) {
+			VERBOSE_MSG("<= lame response: non-auth sent negative response\n");
+			return KR_STATE_FAIL;
+		}
+	}
+
+	const knot_dname_t *cname = NULL;
+	/* Process answer type */
+	int state = unroll_cname(pkt, req, false, &cname);
+	if (state != kr_ok()) {
+		return state;
+	}
 	/* Make sure that this is an authoritative answer (even with AA=0) for other layers */
 	knot_wire_set_aa(pkt->wire);
 	/* Either way it resolves current query. */
@@ -454,61 +552,104 @@ static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
 	/* Follow canonical name as next SNAME. */
 	if (!knot_dname_is_equal(cname, query->sname)) {
 		/* Check if target record has been already copied */
+		query->flags |= QUERY_CNAME;
 		if (is_final) {
-			const knot_pktsection_t *an = knot_pkt_section(req->answer, KNOT_ANSWER);
-			for (unsigned i = 0; i < an->count; ++i) {
-				const knot_rrset_t *rr = knot_pkt_rr(an, i);
-				if (!knot_dname_is_equal(rr->owner, cname)) {
-					continue;
-				}
-				if ((rr->rclass != query->sclass) ||
-				    (rr->type != query->stype)) {
-					continue;
-				}
-				finalize_answer(pkt, query, req);
-				return KNOT_STATE_DONE;
+			state = process_final(pkt, req, cname);
+			if (state != kr_ok()) {
+				return state;
 			}
 		}
-		DEBUG_MSG("<= cname chain, following\n");
+		VERBOSE_MSG("<= cname chain, following\n");
 		/* Check if the same query was already resolved */
 		for (int i = 0; i < req->rplan.resolved.len; ++i) {
 			struct kr_query * q = req->rplan.resolved.at[i];
-			if (q->sclass == query->sclass &&
+			if (q->parent == query->parent &&
+			    q->sclass == query->sclass &&
 			    q->stype == query->stype   &&
 			    knot_dname_is_equal(q->sname, cname)) {
-				DEBUG_MSG("<= cname chain loop\n");
-				return KNOT_STATE_FAIL;
+				VERBOSE_MSG("<= cname chain loop\n");
+				return KR_STATE_FAIL;
 			}
 		}
 		struct kr_query *next = kr_rplan_push(&req->rplan, query->parent, cname, query->sclass, query->stype);
 		if (!next) {
-			return KNOT_STATE_FAIL;
+			return KR_STATE_FAIL;
 		}
 		next->flags |= QUERY_AWAIT_CUT;
-		/* Want DNSSEC if it's posible to secure this name (e.g. is covered by any TA) */
-		if (kr_ta_covers(&req->ctx->trust_anchors, cname) &&
+		if (query->flags & QUERY_DNSSEC_INSECURE) {
+			next->flags &= ~QUERY_DNSSEC_WANT;
+			next->flags |= QUERY_DNSSEC_INSECURE;
+		} else if (kr_ta_covers(&req->ctx->trust_anchors, cname) &&
 		    !kr_ta_covers(&req->ctx->negative_anchors, cname)) {
+			/* Want DNSSEC if it's posible to secure
+			 * this name (e.g. is covered by any TA) */
 			next->flags |= QUERY_DNSSEC_WANT;
 		}
+		state = pick_authority(pkt, req, false);
+		if (state != kr_ok()) {
+			return KR_STATE_FAIL;
+		}
 	} else if (!query->parent) {
+		const bool to_wire = ((pkt_class & (PKT_NXDOMAIN|PKT_NODATA)) != 0);
+		state = pick_authority(pkt, req, to_wire);
+		if (state != kr_ok()) {
+			return KR_STATE_FAIL;
+		}
 		finalize_answer(pkt, query, req);
+	} else {
+		state = pick_authority(pkt, req, false);
+		if (state != kr_ok()) {
+			return KR_STATE_FAIL;
+		}
 	}
-	return KNOT_STATE_DONE;
+	return KR_STATE_DONE;
 }
 
+/** @internal like process_answer() but for the forwarding mode. */
+static int process_stub(knot_pkt_t *pkt, struct kr_request *req)
+{
+	struct kr_query *query = req->current_query;
+	assert(query->flags & QUERY_STUB);
+	/* Pick all answer RRs. */
+	const knot_pktsection_t *an = knot_pkt_section(pkt, KNOT_ANSWER);
+	for (unsigned i = 0; i < an->count; ++i) {
+		const knot_rrset_t *rr = knot_pkt_rr(an, i);
+		int err = kr_ranked_rrarray_add(&req->answ_selected, rr,
+			      KR_VLDRANK_INITIAL, true, query->uid, &req->pool);
+		if (err != kr_ok()) {
+			return KR_STATE_FAIL;
+		}
+	}
+
+	knot_wire_set_aa(pkt->wire);
+	query->flags |= QUERY_RESOLVED;
+
+	/* Pick authority RRs. */
+	int pkt_class = kr_response_classify(pkt);
+	const bool to_wire = ((pkt_class & (PKT_NXDOMAIN|PKT_NODATA)) != 0);
+	int err = pick_authority(pkt, req, to_wire);
+	if (err != kr_ok()) {
+		return KR_STATE_FAIL;
+	}
+
+	finalize_answer(pkt, query, req);
+	return KR_STATE_DONE;
+}
+
+
 /** Error handling, RFC1034 5.3.3, 4d. */
 static int resolve_error(knot_pkt_t *pkt, struct kr_request *req)
 {
-	return KNOT_STATE_FAIL;
+	return KR_STATE_FAIL;
 }
 
 /* State-less single resolution iteration step, not needed. */
-static int reset(knot_layer_t *ctx)  { return KNOT_STATE_PRODUCE; }
+static int reset(kr_layer_t *ctx)  { return KR_STATE_PRODUCE; }
 
 /* Set resolution context and parameters. */
-static int begin(knot_layer_t *ctx, void *module_param)
+static int begin(kr_layer_t *ctx)
 {
-	if (ctx->state & (KNOT_STATE_DONE|KNOT_STATE_FAIL)) {
+	if (ctx->state & (KR_STATE_DONE|KR_STATE_FAIL)) {
 		return ctx->state;
 	}
 	/*
@@ -518,11 +659,20 @@ static int begin(knot_layer_t *ctx, void *module_param)
 	 * Server cookie queries must be handled by the cookie module/layer
 	 * before this layer.
 	 */
-	const struct kr_request *req = ctx->data;
-	const knot_pkt_t *pkt = req->qsource.packet;
+	const knot_pkt_t *pkt = ctx->req->qsource.packet;
 	if (!pkt || knot_wire_get_qdcount(pkt->wire) == 0) {
-		return KNOT_STATE_FAIL;
+		return KR_STATE_FAIL;
 	}
+
+	struct kr_query *qry = ctx->req->current_query;
+	/* Avoid any other classes, and avoid any meta-types ~~except for ANY~~. */
+	if (qry->sclass != KNOT_CLASS_IN
+	    || (knot_rrtype_is_metatype(qry->stype)
+		    /* && qry->stype != KNOT_RRTYPE_ANY hmm ANY seems broken ATM */)) {
+		knot_wire_set_rcode(ctx->req->answer->wire, KNOT_RCODE_NOTIMPL);
+		return KR_STATE_FAIL;
+	}
+
 	return reset(ctx);
 }
 
@@ -543,36 +693,50 @@ int kr_make_query(struct kr_query *query, knot_pkt_t *pkt)
 	query->id = kr_rand_uint(UINT16_MAX);
 	knot_wire_set_id(pkt->wire, query->id);
 	pkt->parsed = pkt->size;
+	WITH_VERBOSE {
+		char name_str[KNOT_DNAME_MAXLEN], type_str[16];
+		knot_dname_to_str(name_str, query->sname, sizeof(name_str));
+		knot_rrtype_to_string(query->stype, type_str, sizeof(type_str));
+		QVERBOSE_MSG(query, "'%s' type '%s' id was assigned, parent id %hu\n",
+			    name_str, type_str, query->parent ? query->parent->id : 0);
+	}
 	return kr_ok();
 }
 
-static int prepare_query(knot_layer_t *ctx, knot_pkt_t *pkt)
+static int prepare_query(kr_layer_t *ctx, knot_pkt_t *pkt)
 {
 	assert(pkt && ctx);
-	struct kr_request *req = ctx->data;
+	struct kr_request *req = ctx->req;
 	struct kr_query *query = req->current_query;
-	if (!query || ctx->state & (KNOT_STATE_DONE|KNOT_STATE_FAIL)) {
+	if (!query || ctx->state & (KR_STATE_DONE|KR_STATE_FAIL)) {
 		return ctx->state;
 	}
 
 	/* Make query */
 	int ret = kr_make_query(query, pkt);
 	if (ret != 0) {
-		return KNOT_STATE_FAIL;
+		return KR_STATE_FAIL;
 	}
 
-	return KNOT_STATE_CONSUME;
+	query->uid = req->rplan.next_uid;
+	req->rplan.next_uid += 1;
+
+	return KR_STATE_CONSUME;
 }
 
 static int resolve_badmsg(knot_pkt_t *pkt, struct kr_request *req, struct kr_query *query)
 {
+
 #ifndef STRICT_MODE
 	/* Work around broken auths/load balancers */
 	if (query->flags & QUERY_SAFEMODE) {
 		return resolve_error(pkt, req);
-	} else {
+	} else if (query->flags & QUERY_NO_MINIMIZE) {
 		query->flags |= QUERY_SAFEMODE;
-		return KNOT_STATE_DONE;
+		return KR_STATE_DONE;
+	} else {
+		query->flags |= QUERY_NO_MINIMIZE;
+		return KR_STATE_DONE;
 	}
 #else
 		return resolve_error(pkt, req);
@@ -583,41 +747,58 @@ static int resolve_badmsg(knot_pkt_t *pkt, struct kr_request *req, struct kr_que
  *
  *  This roughly corresponds to RFC1034, 5.3.3 4a-d.
  */
-static int resolve(knot_layer_t *ctx, knot_pkt_t *pkt)
+static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 {
 	assert(pkt && ctx);
-	struct kr_request *req = ctx->data;
+	struct kr_request *req = ctx->req;
 	struct kr_query *query = req->current_query;
-	if (!query || (query->flags & (QUERY_RESOLVED|QUERY_BADCOOKIE_AGAIN))) {
+	if (!query) {
+		return ctx->state;
+	}
+
+	WITH_VERBOSE {
+	if (query->flags & QUERY_TRACE) {
+		VERBOSE_MSG("<= answer received:\n");
+		kr_pkt_print(pkt);
+	}
+	}
+
+	if (query->flags & (QUERY_RESOLVED|QUERY_BADCOOKIE_AGAIN)) {
 		return ctx->state;
 	}
 
 	/* Check for packet processing errors first.
 	 * Note - we *MUST* check if it has at least a QUESTION,
 	 * otherwise it would crash on accessing QNAME. */
-	if (pkt->parsed < pkt->size || pkt->parsed <= KNOT_WIRE_HEADER_SIZE) {
-		DEBUG_MSG("<= malformed response\n");
+#ifdef STRICT_MODE
+	if (pkt->parsed < pkt->size) {
+		VERBOSE_MSG("<= pkt contains excessive data\n");
+		return resolve_badmsg(pkt, req, query);
+	} else
+#endif
+	if (pkt->parsed <= KNOT_WIRE_HEADER_SIZE) {
+		VERBOSE_MSG("<= malformed response\n");
 		return resolve_badmsg(pkt, req, query);
 	} else if (!is_paired_to_query(pkt, query)) {
-		DEBUG_MSG("<= ignoring mismatching response\n");
+		VERBOSE_MSG("<= ignoring mismatching response\n");
 		/* Force TCP, to work around authoritatives messing up question
 		 * without yielding to spoofed responses. */
 		query->flags |= QUERY_TCP;
 		return resolve_badmsg(pkt, req, query);
 	} else if (knot_wire_get_tc(pkt->wire)) {
-		DEBUG_MSG("<= truncated response, failover to TCP\n");
+		VERBOSE_MSG("<= truncated response, failover to TCP\n");
 		if (query) {
 			/* Fail if already on TCP. */
 			if (query->flags & QUERY_TCP) {
-				DEBUG_MSG("<= TC=1 with TCP, bailing out\n");
+				VERBOSE_MSG("<= TC=1 with TCP, bailing out\n");
 				return resolve_error(pkt, req);
 			}
 			query->flags |= QUERY_TCP;
 		}
-		return KNOT_STATE_CONSUME;
+		return KR_STATE_CONSUME;
 	}
 
-#ifndef NDEBUG
+#ifndef NOVERBOSELOG
 	const knot_lookup_t *rcode = knot_lookup_by_id(knot_rcode_names, knot_wire_get_rcode(pkt->wire));
 #endif
 
@@ -628,34 +809,41 @@ static int resolve(knot_layer_t *ctx, knot_pkt_t *pkt)
 		break; /* OK */
 	case KNOT_RCODE_REFUSED:
 	case KNOT_RCODE_SERVFAIL: {
-		DEBUG_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
+		if (query->flags & QUERY_STUB) { break; } /* Pass through in stub mode */
+		VERBOSE_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
 		query->fails += 1;
 		if (query->fails >= KR_QUERY_NSRETRY_LIMIT) {
 			query->fails = 0; /* Reset per-query counter. */
 			return resolve_error(pkt, req);
 		} else {
 			query->flags |= QUERY_NO_MINIMIZE; /* Drop minimisation as a safe-guard. */
-			return KNOT_STATE_CONSUME;
+			return KR_STATE_CONSUME;
 		}
 	}
 	case KNOT_RCODE_FORMERR:
 	case KNOT_RCODE_NOTIMPL:
-		DEBUG_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
+		VERBOSE_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
 		return resolve_badmsg(pkt, req, query);
 	default:
-		DEBUG_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
+		VERBOSE_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
 		return resolve_error(pkt, req);
 	}
 
+	/* Forwarding/stub mode is special. */
+	if (query->flags & QUERY_STUB) {
+		return process_stub(pkt, req);
+	}
+
 	/* Resolve authority to see if it's referral or authoritative. */
 	int state = process_authority(pkt, req);
 	switch(state) {
-	case KNOT_STATE_CONSUME: /* Not referral, process answer. */
-		DEBUG_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
+	case KR_STATE_CONSUME: /* Not referral, process answer. */
+		VERBOSE_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
 		state = process_answer(pkt, req);
 		break;
-	case KNOT_STATE_DONE: /* Referral */
-		DEBUG_MSG("<= referral response, follow\n");
+	case KR_STATE_DONE: /* Referral */
+		state = process_referral_answer(pkt,req);
+		VERBOSE_MSG("<= referral response, follow\n");
 		break;
 	default:
 		break;
@@ -665,9 +853,9 @@ static int resolve(knot_layer_t *ctx, knot_pkt_t *pkt)
 }
 
 /** Module implementation. */
-const knot_layer_api_t *iterate_layer(struct kr_module *module)
+const kr_layer_api_t *iterate_layer(struct kr_module *module)
 {
-	static const knot_layer_api_t _layer = {
+	static const kr_layer_api_t _layer = {
 		.begin = &begin,
 		.reset = &reset,
 		.consume = &resolve,
@@ -678,4 +866,4 @@ const knot_layer_api_t *iterate_layer(struct kr_module *module)
 
 KR_MODULE_EXPORT(iterate)
 
-#undef DEBUG_MSG
+#undef VERBOSE_MSG
