@@ -1,4 +1,4 @@
-/*  Copyright (C) 2014 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -27,17 +27,19 @@
 #include "contrib/ucw/lib.h"
 
 /** Some built-in unfairness ... */
+#ifndef FAVOUR_IPV6
 #define FAVOUR_IPV6 20 /* 20ms bonus for v6 */
+#endif
 
 /** @internal Macro to set address structure. */
-#define ADDR_SET(sa, family, addr, len) do {\
+#define ADDR_SET(sa, family, addr, len, port) do {\
     	memcpy(&sa ## _addr, (addr), (len)); \
     	sa ## _family = (family); \
-	sa ## _port = htons(KR_DNS_PORT); \
+	sa ## _port = htons(port); \
 } while (0)
 
 /** Update nameserver representation with current name/address pair. */
-static void update_nsrep(struct kr_nsrep *ns, size_t pos, uint8_t *addr, size_t addr_len)
+static void update_nsrep(struct kr_nsrep *ns, size_t pos, uint8_t *addr, size_t addr_len, int port)
 {
 	if (addr == NULL) {
 		ns->addr[pos].ip.sa_family = AF_UNSPEC;
@@ -49,9 +51,9 @@ static void update_nsrep(struct kr_nsrep *ns, size_t pos, uint8_t *addr, size_t
 
 	switch(addr_len) {
 	case sizeof(struct in_addr):
-		ADDR_SET(ns->addr[pos].ip4.sin, AF_INET, addr, addr_len); break;
+		ADDR_SET(ns->addr[pos].ip4.sin, AF_INET, addr, addr_len, port); break;
 	case sizeof(struct in6_addr):
-		ADDR_SET(ns->addr[pos].ip6.sin6, AF_INET6, addr, addr_len); break;
+		ADDR_SET(ns->addr[pos].ip6.sin6, AF_INET6, addr, addr_len, port); break;
 	default: assert(0); break;
 	}
 }
@@ -69,7 +71,7 @@ static void update_nsrep_set(struct kr_nsrep *ns, const knot_dname_t *name, uint
 		if (addr[i]) {
 			void *addr_val = pack_obj_val(addr[i]);
 			size_t len = pack_obj_len(addr[i]);
-			update_nsrep(ns, i, addr_val, len);
+			update_nsrep(ns, i, addr_val, len, KR_DNS_PORT);
 		} else {
 			break;
 		}
@@ -96,7 +98,7 @@ static unsigned eval_addr_set(pack_t *addr_set, kr_nsrep_lru_t *rttcache, unsign
 		}
 		/* Get RTT for this address (if known) */
 		if (is_valid) {
-			unsigned *cached = rttcache ? lru_get(rttcache, val, len) : NULL;
+			unsigned *cached = rttcache ? lru_get_try(rttcache, val, len) : NULL;
 			unsigned addr_score = (cached) ? *cached : KR_NS_GLUED;
 			if (addr_score < score + favour) {
 				/* Shake down previous contenders */
@@ -122,7 +124,8 @@ static int eval_nsrep(const char *k, void *v, void *baton)
 
 	/* Fetch NS reputation */
 	if (ctx->cache_rep) {
-		unsigned *cached = lru_get(ctx->cache_rep, k, knot_dname_size((const uint8_t *)k));
+		unsigned *cached = lru_get_try(ctx->cache_rep, k,
+					knot_dname_size((const uint8_t *)k));
 		if (cached) {
 			reputation = *cached;
 		}
@@ -169,16 +172,31 @@ static int eval_nsrep(const char *k, void *v, void *baton)
 	return kr_ok();
 }
 
-int kr_nsrep_set(struct kr_query *qry, uint8_t *addr, size_t addr_len)
+int kr_nsrep_set(struct kr_query *qry, size_t index, uint8_t *addr, size_t addr_len, int port)
 {
-	if (!qry || !addr) {
+	if (!qry) {
 		return kr_error(EINVAL);
 	}
+	if (index >= KR_NSREP_MAXADDR) {
+		return kr_error(ENOSPC);
+	}
 	qry->ns.name = (const uint8_t *)"";
-	qry->ns.score = KR_NS_UNKNOWN;
-	qry->ns.reputation = 0;
-	update_nsrep(&qry->ns, 0, addr, addr_len);
-	update_nsrep(&qry->ns, 1, NULL, 0);
+	/* Reset score on first entry */
+	if (index == 0) {
+		qry->ns.score = KR_NS_UNKNOWN;
+		qry->ns.reputation = 0;
+	}
+	/* Retrieve RTT from cache */
+	if (addr && addr_len > 0) {
+		struct kr_context *ctx = qry->ns.ctx;
+		unsigned *score = ctx
+			? lru_get_try(ctx->cache_rtt, (const char *)addr, addr_len)
+			: NULL;
+		if (score) {
+			qry->ns.score = MIN(qry->ns.score, *score);
+		}
+	}
+	update_nsrep(&qry->ns, index, addr, addr_len, port);
 	return kr_ok();
 }
 
@@ -240,9 +258,9 @@ int kr_nsrep_update_rtt(struct kr_nsrep *ns, const struct sockaddr *addr,
 			addr_len = sizeof(struct in6_addr);
 		}
 	}
-	unsigned *cur = lru_set(cache, addr_in, addr_len);
+	unsigned *cur = lru_get_new(cache, addr_in, addr_len);
 	if (!cur) {
-		return kr_error(ENOMEM);
+		return kr_ok();
 	}
 	/* Score limits */
 	if (score > KR_NS_MAX_SCORE) {
@@ -260,6 +278,7 @@ int kr_nsrep_update_rtt(struct kr_nsrep *ns, const struct sockaddr *addr,
 	case KR_NS_UPDATE: *cur = (*cur + score) / 2; break;
 	case KR_NS_RESET:  *cur = score; break;
 	case KR_NS_ADD:    *cur = MIN(KR_NS_MAX_SCORE - 1, *cur + score); break;
+	case KR_NS_MAX:    *cur = MAX(*cur, score); break;
 	default: break;
 	}
 	return kr_ok();
@@ -274,10 +293,9 @@ int kr_nsrep_update_rep(struct kr_nsrep *ns, unsigned reputation, kr_nsrep_lru_t
 	/* Store in the struct */
 	ns->reputation = reputation;
 	/* Store reputation in the LRU cache */
-	unsigned *cur = lru_set(cache, (const char *)ns->name, knot_dname_size(ns->name));
-	if (!cur) {
-		return kr_error(ENOMEM);
+	unsigned *cur = lru_get_new(cache, (const char *)ns->name, knot_dname_size(ns->name));
+	if (cur) {
+		*cur = reputation;
 	}
-	*cur = reputation;
 	return kr_ok();
 }
