@@ -1,4 +1,4 @@
-/*  Copyright (C) 2014 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -35,15 +35,26 @@
 #define KNOT_EDNS_OPTION_COOKIE 10
 #endif /* defined(ENABLE_COOKIES) */
 
-#define DEBUG_MSG(qry, fmt...) QRDEBUG((qry), "resl",  fmt)
+#define VERBOSE_MSG(qry, fmt...) QRVERBOSE((qry), "resl",  fmt)
+
+/** @internal Set @a yielded to all RRs with matching @a qry_uid. */
+static void set_yield(ranked_rr_array_t *array, const uint32_t qry_uid, const bool yielded)
+{
+	for (unsigned i = 0; i < array->len; ++i) {
+		ranked_rr_array_entry_t *entry = array->at[i];
+		if (entry->qry_uid == qry_uid) {
+			entry->yielded = yielded;
+		}
+	}
+}
 
 /**
  * @internal Defer execution of current query.
  * The current layer state and input will be pushed to a stack and resumed on next iteration.
  */
-static int consume_yield(knot_layer_t *ctx, knot_pkt_t *pkt)
+static int consume_yield(kr_layer_t *ctx, knot_pkt_t *pkt)
 {
-	struct kr_request *req = ctx->data;
+	struct kr_request *req = ctx->req;
 	knot_pkt_t *pkt_copy = knot_pkt_new(NULL, pkt->size, &req->pool);
 	struct kr_layer_pickle *pickle = mm_alloc(&req->pool, sizeof(*pickle));
 	if (pickle && pkt_copy && knot_pkt_copy(pkt_copy, pkt) == 0) {
@@ -53,38 +64,40 @@ static int consume_yield(knot_layer_t *ctx, knot_pkt_t *pkt)
 		pickle->pkt = pkt_copy;
 		pickle->next = qry->deferred;
 		qry->deferred = pickle;
+		set_yield(&req->answ_selected, qry->uid, true);
+		set_yield(&req->auth_selected, qry->uid, true);
 		return kr_ok();
 	}
 	return kr_error(ENOMEM);
 }
-static int begin_yield(knot_layer_t *ctx, void *module) { return kr_ok(); }
-static int reset_yield(knot_layer_t *ctx) { return kr_ok(); }
-static int finish_yield(knot_layer_t *ctx) { return kr_ok(); }
-static int produce_yield(knot_layer_t *ctx, knot_pkt_t *pkt) { return kr_ok(); }
+static int begin_yield(kr_layer_t *ctx) { return kr_ok(); }
+static int reset_yield(kr_layer_t *ctx) { return kr_ok(); }
+static int finish_yield(kr_layer_t *ctx) { return kr_ok(); }
+static int produce_yield(kr_layer_t *ctx, knot_pkt_t *pkt) { return kr_ok(); }
 
 /** @internal Macro for iterating module layers. */
-#define RESUME_LAYERS(from, req, qry, func, ...) \
-    (req)->current_query = (qry); \
-	for (size_t i = (from); i < (req)->ctx->modules->len; ++i) { \
-		struct kr_module *mod = (req)->ctx->modules->at[i]; \
+#define RESUME_LAYERS(from, r, qry, func, ...) \
+    (r)->current_query = (qry); \
+	for (size_t i = (from); i < (r)->ctx->modules->len; ++i) { \
+		struct kr_module *mod = (r)->ctx->modules->at[i]; \
 		if (mod->layer) { \
-			struct knot_layer layer = {.state = (req)->state, .api = mod->layer(mod), .data = (req)}; \
+			struct kr_layer layer = {.state = (r)->state, .api = mod->layer(mod), .req = (r)}; \
 			if (layer.api && layer.api->func) { \
-				(req)->state = layer.api->func(&layer, ##__VA_ARGS__); \
-				if ((req)->state == KNOT_STATE_YIELD) { \
+				(r)->state = layer.api->func(&layer, ##__VA_ARGS__); \
+				if ((r)->state == KR_STATE_YIELD) { \
 					func ## _yield(&layer, ##__VA_ARGS__); \
 					break; \
 				} \
 			} \
 		} \
 	} /* Invalidate current query. */ \
-	(req)->current_query = NULL
+	(r)->current_query = NULL
 
 /** @internal Macro for starting module iteration. */
 #define ITERATE_LAYERS(req, qry, func, ...) RESUME_LAYERS(0, req, qry, func, ##__VA_ARGS__)
 
 /** @internal Find layer id matching API. */
-static inline size_t layer_id(struct kr_request *req, const struct knot_layer_api *api) {
+static inline size_t layer_id(struct kr_request *req, const struct kr_layer_api *api) {
 	module_array_t *modules = req->ctx->modules;
 	for (size_t i = 0; i < modules->len; ++i) {
 		struct kr_module *mod = modules->at[i];
@@ -164,37 +177,104 @@ static void check_empty_nonterms(struct kr_query *qry, knot_pkt_t *pkt, struct k
 	}
 }
 
-static int ns_fetch_cut(struct kr_query *qry, struct kr_request *req, knot_pkt_t *pkt)
+static int ns_fetch_cut(struct kr_query *qry, const knot_dname_t *requested_name,
+			struct kr_request *req, knot_pkt_t *pkt)
 {
-	int ret = 0;
+	map_t *trust_anchors = &req->ctx->trust_anchors;
+	map_t *negative_anchors = &req->ctx->negative_anchors;
 
-	/* Find closest zone cut from cache */
-	struct kr_cache *cache = &req->ctx->cache;
-	if (kr_cache_is_open(cache)) {
-		/* If at/subdomain of parent zone cut, start from its encloser.
-		 * This is for case when we get to a dead end (and need glue from parent), or DS refetch. */
-		struct kr_query *parent = qry->parent;
-		bool secured = (qry->flags & QUERY_DNSSEC_WANT);
-		if (parent && parent->zone_cut.name[0] != '\0' && knot_dname_in(parent->zone_cut.name, qry->sname)) {
-			const knot_dname_t *encloser = knot_wire_next_label(parent->zone_cut.name, NULL);
-			ret = kr_zonecut_find_cached(req->ctx, &qry->zone_cut, encloser, qry->timestamp.tv_sec, &secured);
-		} else {
-			ret = kr_zonecut_find_cached(req->ctx, &qry->zone_cut, qry->sname, qry->timestamp.tv_sec, &secured);
-		}
-		/* Check if there's a non-terminal between target and current cut. */
-		if (ret == 0) {
-			check_empty_nonterms(qry, pkt, cache, qry->timestamp.tv_sec);
-			/* Go insecure if the zone cut is provably insecure */
-			if ((qry->flags & QUERY_DNSSEC_WANT) && !secured) {
-				DEBUG_MSG(qry, "=> NS is provably without DS, going insecure\n");
-				qry->flags &= ~QUERY_DNSSEC_WANT;
-				qry->flags |= QUERY_DNSSEC_INSECURE;
-			}
-		}
+	/* It can occur that here parent query already have
+	 * provably insecured zonecut which not in the cache yet. */
+	const bool is_insecured = ((qry->parent != NULL) &&
+				   (qry->parent->flags & QUERY_DNSSEC_INSECURE) != 0);
+
+	/* Want DNSSEC if it's possible to secure this name
+	 * (e.g. is covered by any TA) */
+	if (is_insecured) {
+		/* If parent is unsecured we don't want DNSSEC
+		 * even if cut name is covered by TA. */
+		qry->flags &= ~QUERY_DNSSEC_WANT;
+		qry->flags |= QUERY_DNSSEC_INSECURE;
+	} else if (!kr_ta_covers(negative_anchors, qry->zone_cut.name) &&
+		   kr_ta_covers(trust_anchors, qry->zone_cut.name)) {
+		qry->flags |= QUERY_DNSSEC_WANT;
 	} else {
-		ret = kr_error(ENOENT);
+		qry->flags &= ~QUERY_DNSSEC_WANT;
 	}
-	return ret;
+
+	struct kr_zonecut cut_found = {0};
+	kr_zonecut_init(&cut_found, requested_name, req->rplan.pool);
+	/* Cut that has been found can differs from cut that has been requested.
+	 * So if not already insecured,
+	 * try to fetch ta & keys even if initial cut name not covered by TA */
+	bool secured = !is_insecured;
+	int ret = kr_zonecut_find_cached(req->ctx, &cut_found, requested_name,
+					 qry->timestamp.tv_sec, &secured);
+	if (ret == kr_error(ENOENT)) {
+		/* No cached cut found, start from SBELT
+		 * and issue priming query. */
+		ret = kr_zonecut_set_sbelt(req->ctx, &qry->zone_cut);
+		if (ret != 0) {
+			return KR_STATE_FAIL;
+		}
+		VERBOSE_MSG(qry, "=> using root hints\n");
+		qry->flags &= ~QUERY_AWAIT_CUT;
+		kr_zonecut_deinit(&cut_found);
+		return KR_STATE_DONE;
+	} else if (ret != kr_ok()) {
+		kr_zonecut_deinit(&cut_found);
+		return KR_STATE_FAIL;
+	}
+
+	/* Find out security status.
+	 * Go insecure if the zone cut is provably insecure */
+	if ((qry->flags & QUERY_DNSSEC_WANT) && !secured) {
+		VERBOSE_MSG(qry, "=> NS is provably without DS, going insecure\n");
+		qry->flags &= ~QUERY_DNSSEC_WANT;
+		qry->flags |= QUERY_DNSSEC_INSECURE;
+	}
+	/* Zonecut name can change, check it again
+	 * to prevent unnecessary DS & DNSKEY queries */
+	if (!(qry->flags & QUERY_DNSSEC_INSECURE) &&
+	    !kr_ta_covers(negative_anchors, cut_found.name) &&
+	    kr_ta_covers(trust_anchors, cut_found.name)) {
+		qry->flags |= QUERY_DNSSEC_WANT;
+	} else {
+		qry->flags &= ~QUERY_DNSSEC_WANT;
+	}
+	/* Check if any DNSKEY found for cached cut */
+	if ((qry->flags & QUERY_DNSSEC_WANT) &&
+	    (cut_found.key == NULL)) {
+		/* No DNSKEY was found for cached cut.
+		 * If no glue were fetched for this cut,
+		 * we have got circular dependance - must fetch A\AAAA
+		 * from authoritative, but we have no key to verify it.
+		 * TODO - try to refetch cut only if no glue were fetched */
+		kr_zonecut_deinit(&cut_found);
+		if (requested_name[0] != '\0' ) {
+			/* If not root - try next label */
+			return KR_STATE_CONSUME;
+		}
+		/* No cached cut & keys found, start from SBELT */
+		ret = kr_zonecut_set_sbelt(req->ctx, &qry->zone_cut);
+		if (ret != 0) {
+			return KR_STATE_FAIL;
+		}
+		VERBOSE_MSG(qry, "=> using root hints\n");
+		qry->flags &= ~QUERY_AWAIT_CUT;
+		return KR_STATE_DONE;
+	}
+	/* Copy fetched name */
+	qry->zone_cut.name = knot_dname_copy(cut_found.name, qry->zone_cut.pool);
+	/* Copy fetched address set */
+	kr_zonecut_copy(&qry->zone_cut, &cut_found);
+	/* Copy fetched ta & keys */
+	kr_zonecut_copy_trust(&qry->zone_cut, &cut_found);
+	/* Check if there's a non-terminal between target and current cut. */
+	struct kr_cache *cache = &req->ctx->cache;
+	check_empty_nonterms(qry, pkt, cache, qry->timestamp.tv_sec);
+	/* Cut found */
+	return KR_STATE_PRODUCE;
 }
 
 static int ns_resolve_addr(struct kr_query *qry, struct kr_request *param)
@@ -223,13 +303,13 @@ static int ns_resolve_addr(struct kr_query *qry, struct kr_request *param)
 	if (!next_type || kr_rplan_satisfies(qry->parent, qry->ns.name, KNOT_CLASS_IN, next_type)) {
 		/* Fall back to SBELT if root server query fails. */
 		if (!next_type && qry->zone_cut.name[0] == '\0') {
-			DEBUG_MSG(qry, "=> fallback to root hints\n");
+			VERBOSE_MSG(qry, "=> fallback to root hints\n");
 			kr_zonecut_set_sbelt(ctx, &qry->zone_cut);
 			qry->flags |= QUERY_NO_THROTTLE; /* Pick even bad SBELT servers */
 			return kr_error(EAGAIN);
 		}
 		/* No IPv4 nor IPv6, flag server as unuseable. */
-		DEBUG_MSG(qry, "=> unresolvable NS address, bailing out\n");
+		VERBOSE_MSG(qry, "=> unresolvable NS address, bailing out\n");
 		qry->ns.reputation |= KR_NS_NOIP4 | KR_NS_NOIP6;
 		kr_nsrep_update_rep(&qry->ns, qry->ns.reputation, ctx->cache_rep);
 		invalidate_ns(rplan, qry);
@@ -308,6 +388,9 @@ static int edns_create(knot_pkt_t *pkt, knot_pkt_t *template, struct kr_request
 		wire_size += KR_COOKIE_OPT_MAX_LEN;
 	}
 #endif /* defined(ENABLE_COOKIES) */
+	if (req->has_tls && req->ctx->tls_padding >= 2) {
+		wire_size += KNOT_EDNS_OPTION_HDRLEN + req->ctx->tls_padding;
+	}
 	return knot_pkt_reserve(pkt, wire_size);
 }
 
@@ -337,8 +420,58 @@ static void write_extra_records(rr_array_t *arr, knot_pkt_t *answer)
 	}
 }
 
-static int answer_fail(knot_pkt_t *answer)
+static void write_extra_ranked_records(ranked_rr_array_t *arr, knot_pkt_t *answer)
 {
+	for (size_t i = 0; i < arr->len; ++i) {
+		ranked_rr_array_entry_t * entry = arr->at[i];
+		if (!entry->to_wire) {
+			continue;
+		}
+		knot_rrset_t *rr = entry->rr;
+		if (!knot_pkt_has_dnssec(answer)) {
+			if (rr->type != knot_pkt_qtype(answer) && knot_rrtype_is_dnssec(rr->type)) {
+				continue;
+			}
+		}
+		knot_pkt_put(answer, 0, rr, 0);
+	}
+}
+
+/** @internal Add an EDNS padding RR into the answer if requested and required. */
+static int answer_padding(struct kr_request *request)
+{
+	if (!request || !request->answer || !request->ctx) {
+		assert(false);
+		return kr_error(EINVAL);
+	}
+	uint16_t padding = request->ctx->tls_padding;
+	knot_pkt_t *answer = request->answer;
+	knot_rrset_t *opt_rr = answer->opt_rr;
+
+	if (padding < 2) {
+		return kr_ok();
+	}
+	int32_t max_pad_bytes = knot_edns_get_payload(opt_rr) - (answer->size + knot_rrset_size(opt_rr));
+
+	int32_t pad_bytes = MIN(knot_edns_alignment_size(answer->size, knot_rrset_size(opt_rr), padding),
+				max_pad_bytes);
+
+	if (pad_bytes >= 0) {
+		uint8_t zeros[MAX(1, pad_bytes)];
+		memset(zeros, 0, sizeof(zeros));
+		int r = knot_edns_add_option(opt_rr, KNOT_EDNS_OPTION_PADDING,
+					     pad_bytes, zeros, &answer->mm);
+		if (r != KNOT_EOK) {
+			knot_rrset_clear(opt_rr, &answer->mm);
+			return kr_error(r);
+		}
+	}
+	return kr_ok();
+}
+
+static int answer_fail(struct kr_request *request)
+{
+	knot_pkt_t *answer = request->answer;
 	int ret = kr_pkt_clear_payload(answer);
 	knot_wire_clear_ad(answer->wire);
 	knot_wire_clear_aa(answer->wire);
@@ -346,6 +479,7 @@ static int answer_fail(knot_pkt_t *answer)
 	if (ret == 0 && answer->opt_rr) {
 		/* OPT in SERVFAIL response is still useful for cookies/additional info. */
 		knot_pkt_begin(answer, KNOT_ADDITIONAL);
+		answer_padding(request); /* Ignore failed padding in SERVFAIL answer. */
 		ret = edns_put(answer);
 	}
 	return ret;
@@ -357,24 +491,38 @@ static int answer_finalize(struct kr_request *request, int state)
 	knot_pkt_t *answer = request->answer;
 
 	/* Always set SERVFAIL for bogus answers. */
-	if (state == KNOT_STATE_FAIL && rplan->pending.len > 0) {
+	if (state == KR_STATE_FAIL && rplan->pending.len > 0) {
 		struct kr_query *last = array_tail(rplan->pending);
 		if ((last->flags & QUERY_DNSSEC_WANT) && (last->flags & QUERY_DNSSEC_BOGUS)) {
-			return answer_fail(answer);
+			return answer_fail(request);
 		}
 	}
 
+	if (request->answ_selected.len > 0) {
+		assert(answer->current <= KNOT_ANSWER);
+		/* Write answer records. */
+		if (answer->current < KNOT_ANSWER) {
+			knot_pkt_begin(answer, KNOT_ANSWER);
+		}
+		write_extra_ranked_records(&request->answ_selected, answer);
+	}
+
 	/* Write authority records. */
 	if (answer->current < KNOT_AUTHORITY) {
 		knot_pkt_begin(answer, KNOT_AUTHORITY);
 	}
-	write_extra_records(&request->authority, answer);
+	write_extra_ranked_records(&request->auth_selected, answer);
 	/* Write additional records. */
 	knot_pkt_begin(answer, KNOT_ADDITIONAL);
 	write_extra_records(&request->additional, answer);
 	/* Write EDNS information */
 	int ret = 0;
 	if (answer->opt_rr) {
+		if (request->has_tls) {
+			if (answer_padding(request) != kr_ok()) {
+				return answer_fail(request);
+			}
+		}
 		knot_pkt_begin(answer, KNOT_ADDITIONAL);
 		ret = edns_put(answer);
 	}
@@ -382,12 +530,14 @@ static int answer_finalize(struct kr_request *request, int state)
 	/* Set AD=1 if succeeded and requested secured answer. */
 	const bool has_ad = knot_wire_get_ad(answer->wire);
 	knot_wire_clear_ad(answer->wire);
-	if (state == KNOT_STATE_DONE && rplan->resolved.len > 0) {
+	if (state == KR_STATE_DONE && rplan->resolved.len > 0) {
 		struct kr_query *last = array_tail(rplan->resolved);
 		/* Do not set AD for RRSIG query, as we can't validate it. */
 		const bool secure = (last->flags & QUERY_DNSSEC_WANT) &&
 		                   !(last->flags & QUERY_DNSSEC_INSECURE);
-		if (has_ad && secure && knot_pkt_qtype(answer) != KNOT_RRTYPE_RRSIG) {
+		if (!(last->flags & QUERY_STUB) /* Never set AD if forwarding. */
+		    && has_ad && secure
+		    && knot_pkt_qtype(answer) != KNOT_RRTYPE_RRSIG) {
 			knot_wire_set_ad(answer->wire);
 		}
 	}
@@ -397,11 +547,6 @@ static int answer_finalize(struct kr_request *request, int state)
 
 static int query_finalize(struct kr_request *request, struct kr_query *qry, knot_pkt_t *pkt)
 {
-	/* Randomize query case (if not in safemode) */
-	qry->secret = (qry->flags & QUERY_SAFEMODE) ? 0 : kr_rand_uint(UINT32_MAX);
-	knot_dname_t *qname_raw = (knot_dname_t *)knot_pkt_qname(pkt);
-	randomized_qname_case(qname_raw, qry->secret);
-
 	int ret = 0;
 	knot_pkt_begin(pkt, KNOT_ADDITIONAL);
 	if (!(qry->flags & QUERY_SAFEMODE)) {
@@ -433,14 +578,17 @@ int kr_resolve_begin(struct kr_request *request, struct kr_context *ctx, knot_pk
 	request->ctx = ctx;
 	request->answer = answer;
 	request->options = ctx->options;
-	request->state = KNOT_STATE_CONSUME;
+	request->state = KR_STATE_CONSUME;
 	request->current_query = NULL;
-	array_init(request->authority);
 	array_init(request->additional);
+	array_init(request->answ_selected);
+	array_init(request->auth_selected);
+	request->answ_validated = false;
+	request->auth_validated = false;
 
 	/* Expect first query */
 	kr_rplan_init(&request->rplan, request, &request->pool);
-	return KNOT_STATE_CONSUME;
+	return KR_STATE_CONSUME;
 }
 
 static int resolve_query(struct kr_request *request, const knot_pkt_t *packet)
@@ -449,6 +597,7 @@ static int resolve_query(struct kr_request *request, const knot_pkt_t *packet)
 	const knot_dname_t *qname = knot_pkt_qname(packet);
 	uint16_t qclass = knot_pkt_qclass(packet);
 	uint16_t qtype = knot_pkt_qtype(packet);
+	bool cd_is_set = knot_wire_get_cd(packet->wire);
 	struct kr_query *qry = NULL;
 
 	if (qname != NULL) {
@@ -460,7 +609,7 @@ static int resolve_query(struct kr_request *request, const knot_pkt_t *packet)
 		qry = kr_rplan_push_empty(rplan, NULL);
 	}
 	if (!qry) {
-		return KNOT_STATE_FAIL;
+		return KR_STATE_FAIL;
 	}
 
 	/* Deferred zone cut lookup for this query. */
@@ -479,49 +628,135 @@ static int resolve_query(struct kr_request *request, const knot_pkt_t *packet)
 	knot_wire_clear_aa(answer->wire);
 	knot_wire_set_ra(answer->wire);
 	knot_wire_set_rcode(answer->wire, KNOT_RCODE_NOERROR);
-	if (qry->flags & QUERY_DNSSEC_WANT) {
+
+	if (cd_is_set) {
+		knot_wire_set_cd(answer->wire);
+	} else if (qry->flags & QUERY_DNSSEC_WANT) {
 		knot_wire_set_ad(answer->wire);
 	}
 
 	/* Expect answer, pop if satisfied immediately */
 	request->qsource.packet = packet;
-	ITERATE_LAYERS(request, qry, begin, request);
+	ITERATE_LAYERS(request, qry, begin);
 	request->qsource.packet = NULL;
-	if (request->state == KNOT_STATE_DONE) {
+	if (request->state == KR_STATE_DONE) {
 		kr_rplan_pop(rplan, qry);
 	}
 	return request->state;
 }
 
+KR_PURE static bool kr_inaddr_equal(const struct sockaddr *a, const struct sockaddr *b)
+{
+	const int a_len = kr_inaddr_len(a);
+	const int b_len = kr_inaddr_len(b);
+	return a_len == b_len && memcmp(kr_inaddr(a), kr_inaddr(b), a_len) == 0;
+}
+
+static void update_nslist_rtt(struct kr_context *ctx, struct kr_query *qry, const struct sockaddr *src)
+{
+	/* Do not track in safe mode. */
+	if (qry->flags & QUERY_SAFEMODE) {
+		return;
+	}
+
+	/* Calculate total resolution time from the time the query was generated. */
+	struct timeval now;
+	gettimeofday(&now, NULL);
+	unsigned elapsed = time_diff(&qry->timestamp, &now);
+
+	/* NSs in the preference list prior to the one who responded will be penalised
+	 * with the RETRY timer interval. This is because we know they didn't respond
+	 * for N retries, so their RTT must be at least N * RETRY.
+	 * The NS in the preference list that responded will have RTT relative to the
+	 * time when the query was sent out, not when it was originated.
+	 */
+	for (size_t i = 0; i < KR_NSREP_MAXADDR; ++i) {
+		const struct sockaddr *addr = &qry->ns.addr[i].ip;
+		if (addr->sa_family == AF_UNSPEC) {
+			break;
+		}
+		/* If this address is the source of the answer, update its RTT */
+		if (kr_inaddr_equal(src, addr)) {
+			kr_nsrep_update_rtt(&qry->ns, addr, elapsed, ctx->cache_rtt, KR_NS_UPDATE);
+			WITH_VERBOSE {
+				char addr_str[INET6_ADDRSTRLEN];
+				inet_ntop(addr->sa_family, kr_inaddr(addr), addr_str, sizeof(addr_str));
+				VERBOSE_MSG(qry, "<= server: '%s' rtt: %ld ms\n", addr_str, elapsed);
+			}
+		} else {
+			/* Response didn't come from this IP, but we know the RTT must be at least
+			 * several RETRY timer tries, e.g. if we have addresses [a, b, c] and we have
+			 * tried [a, b] when the answer from 'a' came after 350ms, then we know
+			 * that 'b' didn't respond for at least 350 - (1 * 300) ms. We can't say that
+			 * its RTT is 50ms, but we can say that its score shouldn't be less than 50. */
+			 kr_nsrep_update_rtt(&qry->ns, addr, elapsed, ctx->cache_rtt, KR_NS_MAX);
+			 WITH_VERBOSE {
+			 	char addr_str[INET6_ADDRSTRLEN];
+			 	inet_ntop(addr->sa_family, kr_inaddr(addr), addr_str, sizeof(addr_str));
+			 	VERBOSE_MSG(qry, "<= server: '%s' rtt: >=%ld ms\n", addr_str, elapsed);
+			 }
+		}
+		/* Subtract query start time from elapsed time */
+		if (elapsed < KR_CONN_RETRY) {
+			break;
+		}
+		elapsed = elapsed - KR_CONN_RETRY;
+	}
+}
+
+static void update_nslist_score(struct kr_request *request, struct kr_query *qry, const struct sockaddr *src, knot_pkt_t *packet)
+{
+	struct kr_context *ctx = request->ctx;
+	/* On sucessful answer, update preference list RTT and penalise timer  */
+	if (request->state != KR_STATE_FAIL) {
+		/* Update RTT information for preference list */
+		update_nslist_rtt(ctx, qry, src);
+		/* Do not complete NS address resolution on soft-fail. */
+		const int rcode = packet ? knot_wire_get_rcode(packet->wire) : 0;
+		if (rcode != KNOT_RCODE_SERVFAIL && rcode != KNOT_RCODE_REFUSED) {
+			qry->flags &= ~(QUERY_AWAIT_IPV6|QUERY_AWAIT_IPV4);
+		} else { /* Penalize SERVFAILs. */
+			kr_nsrep_update_rtt(&qry->ns, src, KR_NS_PENALTY, ctx->cache_rtt, KR_NS_ADD);
+		}
+	/* Penalise resolution failures except validation failures. */
+	} else if (!(qry->flags & QUERY_DNSSEC_BOGUS)) {
+		kr_nsrep_update_rtt(&qry->ns, src, KR_NS_TIMEOUT, ctx->cache_rtt, KR_NS_RESET);
+		WITH_VERBOSE {
+			char addr_str[INET6_ADDRSTRLEN];
+			inet_ntop(src->sa_family, kr_inaddr(src), addr_str, sizeof(addr_str));
+			VERBOSE_MSG(qry, "=> server: '%s' flagged as 'bad'\n", addr_str);
+		}
+	}
+}
+
 int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, knot_pkt_t *packet)
 {
 	struct kr_rplan *rplan = &request->rplan;
-	struct kr_context *ctx = request->ctx;
 
 	/* Empty resolution plan, push packet as the new query */
 	if (packet && kr_rplan_empty(rplan)) {
 		if (answer_prepare(request->answer, packet, request) != 0) {
-			return KNOT_STATE_FAIL;
+			return KR_STATE_FAIL;
 		}
 		return resolve_query(request, packet);
 	}
 
 	/* Different processing for network error */
 	struct kr_query *qry = array_tail(rplan->pending);
-
 	bool tried_tcp = (qry->flags & QUERY_TCP);
 	if (!packet || packet->size == 0) {
-		if (tried_tcp)
-			request->state = KNOT_STATE_FAIL;
-		else
+		if (tried_tcp) {
+			request->state = KR_STATE_FAIL;
+		} else {
 			qry->flags |= QUERY_TCP;
+		}
 	} else {
 		/* Packet cleared, derandomize QNAME. */
 		knot_dname_t *qname_raw = (knot_dname_t *)knot_pkt_qname(packet);
 		if (qname_raw && qry->secret != 0) {
 			randomized_qname_case(qname_raw, qry->secret);
 		}
-		request->state = KNOT_STATE_CONSUME;
+		request->state = KR_STATE_CONSUME;
 		if (qry->flags & QUERY_CACHED) {
 			ITERATE_LAYERS(request, qry, consume, packet);
 		} else {
@@ -539,49 +774,21 @@ int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, k
 
 	/* Track RTT for iterative answers */
 	if (src && !(qry->flags & QUERY_CACHED)) {
-		/* Sucessful answer, track RTT and lift any address resolution requests. */
-		if (request->state != KNOT_STATE_FAIL) {
-			/* Do not track in safe mode. */
-			if (!(qry->flags & QUERY_SAFEMODE)) {
-				struct timeval now;
-				gettimeofday(&now, NULL);
-				kr_nsrep_update_rtt(&qry->ns, src, time_diff(&qry->timestamp, &now), ctx->cache_rtt, KR_NS_UPDATE);
-				WITH_DEBUG {
-					char addr_str[INET6_ADDRSTRLEN];
-					inet_ntop(src->sa_family, kr_inaddr(src), addr_str, sizeof(addr_str));
-					DEBUG_MSG(qry, "<= server: '%s' rtt: %ld ms\n", addr_str, time_diff(&qry->timestamp, &now));
-				}
-			}
-			/* Do not complete NS address resolution on soft-fail. */
-			const int rcode = packet ? knot_wire_get_rcode(packet->wire) : 0;
-			if (rcode != KNOT_RCODE_SERVFAIL && rcode != KNOT_RCODE_REFUSED) {
-				qry->flags &= ~(QUERY_AWAIT_IPV6|QUERY_AWAIT_IPV4);
-			} else { /* Penalize SERVFAILs. */
-				kr_nsrep_update_rtt(&qry->ns, src, KR_NS_PENALTY, ctx->cache_rtt, KR_NS_ADD);
-			}
-		/* Do not penalize validation timeouts. */
-		} else if (!(qry->flags & QUERY_DNSSEC_BOGUS)) {
-			kr_nsrep_update_rtt(&qry->ns, src, KR_NS_TIMEOUT, ctx->cache_rtt, KR_NS_RESET);
-			WITH_DEBUG {
-				char addr_str[INET6_ADDRSTRLEN];
-				inet_ntop(src->sa_family, kr_inaddr(src), addr_str, sizeof(addr_str));
-				DEBUG_MSG(qry, "=> server: '%s' flagged as 'bad'\n", addr_str);
-			}
-		}
+		update_nslist_score(request, qry, src, packet);
 	}
 	/* Resolution failed, invalidate current NS. */
-	if (request->state == KNOT_STATE_FAIL) {
+	if (request->state == KR_STATE_FAIL) {
 		invalidate_ns(rplan, qry);
 		qry->flags &= ~QUERY_RESOLVED;
 	}
 
 	/* Pop query if resolved. */
-	if (request->state == KNOT_STATE_YIELD) {
-		return KNOT_STATE_PRODUCE; /* Requery */
+	if (request->state == KR_STATE_YIELD) {
+		return KR_STATE_PRODUCE; /* Requery */
 	} else if (qry->flags & QUERY_RESOLVED) {
 		kr_rplan_pop(rplan, qry);
 	} else if (!tried_tcp && (qry->flags & QUERY_TCP)) {
-		return KNOT_STATE_PRODUCE; /* Requery over TCP */
+		return KR_STATE_PRODUCE; /* Requery over TCP */
 	} else { /* Clear query flags for next attempt */
 		qry->flags &= ~(QUERY_CACHED|QUERY_TCP);
 	}
@@ -590,10 +797,10 @@ int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, k
 
 	/* Do not finish with bogus answer. */
 	if (qry->flags & QUERY_DNSSEC_BOGUS)  {
-		return KNOT_STATE_FAIL;
+		return KR_STATE_FAIL;
 	}
 
-	return kr_rplan_empty(&request->rplan) ? KNOT_STATE_DONE : KNOT_STATE_PRODUCE;
+	return kr_rplan_empty(&request->rplan) ? KR_STATE_DONE : KR_STATE_PRODUCE;
 }
 
 /** @internal Spawn subrequest in current zone cut (no minimization or lookup). */
@@ -625,18 +832,21 @@ static int trust_chain_check(struct kr_request *request, struct kr_query *qry)
 
 	/* Disable DNSSEC if it enters NTA. */
 	if (kr_ta_get(negative_anchors, qry->zone_cut.name)){
-		DEBUG_MSG(qry, ">< negative TA, going insecure\n");
+		VERBOSE_MSG(qry, ">< negative TA, going insecure\n");
 		qry->flags &= ~QUERY_DNSSEC_WANT;
 	}
 	/* Enable DNSSEC if enters a new island of trust. */
-	bool want_secured = (qry->flags & QUERY_DNSSEC_WANT);
-	if (!want_secured && kr_ta_get(trust_anchors, qry->zone_cut.name)) {
+	bool want_secured = (qry->flags & QUERY_DNSSEC_WANT) &&
+			    !knot_wire_get_cd(request->answer->wire);
+	if (!(qry->flags & QUERY_DNSSEC_WANT) &&
+	    !knot_wire_get_cd(request->answer->wire) &&
+	    kr_ta_get(trust_anchors, qry->zone_cut.name)) {
 		qry->flags |= QUERY_DNSSEC_WANT;
 		want_secured = true;
-		WITH_DEBUG {
+		WITH_VERBOSE {
 		char qname_str[KNOT_DNAME_MAXLEN];
 		knot_dname_to_str(qname_str, qry->zone_cut.name, sizeof(qname_str));
-		DEBUG_MSG(qry, ">< TA: '%s'\n", qname_str);
+		VERBOSE_MSG(qry, ">< TA: '%s'\n", qname_str);
 		}
 	}
 	if (want_secured && !qry->zone_cut.trust_anchor) {
@@ -651,10 +861,10 @@ static int trust_chain_check(struct kr_request *request, struct kr_query *qry)
 		/* @todo we could fetch the information from the parent cut, but we don't remember that now */
 		struct kr_query *next = kr_rplan_push(rplan, qry, qry->zone_cut.name, qry->sclass, KNOT_RRTYPE_DS);
 		if (!next) {
-			return KNOT_STATE_FAIL;
+			return KR_STATE_FAIL;
 		}
 		next->flags |= QUERY_AWAIT_CUT|QUERY_DNSSEC_WANT;
-		return KNOT_STATE_DONE;
+		return KR_STATE_DONE;
 	}
 	/* Try to fetch missing DNSKEY (either missing or above current cut).
 	 * Do not fetch if this is a DNSKEY subrequest to avoid circular dependency. */
@@ -663,59 +873,70 @@ static int trust_chain_check(struct kr_request *request, struct kr_query *qry)
 	if (want_secured && refetch_key && !is_dnskey_subreq) {
 		struct kr_query *next = zone_cut_subreq(rplan, qry, ta_name, KNOT_RRTYPE_DNSKEY);
 		if (!next) {
-			return KNOT_STATE_FAIL;
+			return KR_STATE_FAIL;
 		}
-		return KNOT_STATE_DONE;
+		return KR_STATE_DONE;
 	}
 
-	return KNOT_STATE_PRODUCE;
+	return KR_STATE_PRODUCE;
 }
 
 /** @internal Check current zone cut status and credibility, spawn subrequests if needed. */
 static int zone_cut_check(struct kr_request *request, struct kr_query *qry, knot_pkt_t *packet)
 {
-	map_t *trust_anchors = &request->ctx->trust_anchors;
-	map_t *negative_anchors = &request->ctx->negative_anchors;
-
 	/* Stub mode, just forward and do not solve cut. */
 	if (qry->flags & QUERY_STUB) {
-		return KNOT_STATE_PRODUCE;
+		return KR_STATE_PRODUCE;
+	}
+
+	if (!(qry->flags & QUERY_AWAIT_CUT)) {
+		/* The query was resolved from cache.
+		 * Spawn DS \ DNSKEY requests if needed and exit */
+		return trust_chain_check(request, qry);
 	}
 
 	/* The query wasn't resolved from cache,
 	 * now it's the time to look up closest zone cut from cache. */
-	if (qry->flags & QUERY_AWAIT_CUT) {
-		/* Want DNSSEC if it's posible to secure this name (e.g. is covered by any TA) */
-		if (!kr_ta_covers(negative_anchors, qry->zone_cut.name) &&
-		    kr_ta_covers(trust_anchors, qry->zone_cut.name)) {
-			qry->flags |= QUERY_DNSSEC_WANT;
-		} else {
-			qry->flags &= ~QUERY_DNSSEC_WANT;
-		}
-		int ret = ns_fetch_cut(qry, request, packet);
+	struct kr_cache *cache = &request->ctx->cache;
+	if (!kr_cache_is_open(cache)) {
+		int ret = kr_zonecut_set_sbelt(request->ctx, &qry->zone_cut);
 		if (ret != 0) {
-			/* No cached cut found, start from SBELT and issue priming query. */
-			if (ret == kr_error(ENOENT)) {
-				ret = kr_zonecut_set_sbelt(request->ctx, &qry->zone_cut);
-				if (ret != 0) {
-					return KNOT_STATE_FAIL;
-				}
-				DEBUG_MSG(qry, "=> using root hints\n");
-				qry->flags &= ~QUERY_AWAIT_CUT;
-				return KNOT_STATE_DONE;
-			} else {
-				return KNOT_STATE_FAIL;
-			}
-		}
-		/* Update minimized QNAME if zone cut changed */
-		if (qry->zone_cut.name[0] != '\0' && !(qry->flags & QUERY_NO_MINIMIZE)) {
-			if (kr_make_query(qry, packet) != 0) {
-				return KNOT_STATE_FAIL;
-			}
+			return KR_STATE_FAIL;
 		}
+		VERBOSE_MSG(qry, "=> using root hints\n");
 		qry->flags &= ~QUERY_AWAIT_CUT;
+		return KR_STATE_DONE;
 	}
 
+	const knot_dname_t *requested_name = qry->sname;
+	/* If at/subdomain of parent zone cut, start from its encloser.
+	 * This is for case when we get to a dead end
+	 * (and need glue from parent), or DS refetch. */
+	if (qry->parent) {
+		const knot_dname_t *parent = qry->parent->zone_cut.name;
+		if (parent[0] != '\0' && knot_dname_in(parent, qry->sname)) {
+			requested_name = knot_wire_next_label(parent, NULL);
+		}
+	}
+
+	int state = KR_STATE_FAIL;
+	do {
+		state = ns_fetch_cut(qry, requested_name, request, packet);
+		if (state == KR_STATE_DONE || state == KR_STATE_FAIL) {
+			return state;
+		} else if (state == KR_STATE_CONSUME) {
+			requested_name = knot_wire_next_label(requested_name, NULL);
+		}
+	} while (state == KR_STATE_CONSUME);
+
+	/* Update minimized QNAME if zone cut changed */
+	if (qry->zone_cut.name[0] != '\0' && !(qry->flags & QUERY_NO_MINIMIZE)) {
+		if (kr_make_query(qry, packet) != 0) {
+			return KR_STATE_FAIL;
+		}
+	}
+	qry->flags &= ~QUERY_AWAIT_CUT;
+
 	/* Check trust chain */
 	return trust_chain_check(request, qry);
 }
@@ -727,20 +948,22 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 
 	/* No query left for resolution */
 	if (kr_rplan_empty(rplan)) {
-		return KNOT_STATE_FAIL;
+		return KR_STATE_FAIL;
 	}
 	/* If we have deferred answers, resume them. */
 	struct kr_query *qry = array_tail(rplan->pending);
 	if (qry->deferred != NULL) {
 		/* @todo: Refactoring validator, check trust chain before resuming. */
 		switch(trust_chain_check(request, qry)) {
-		case KNOT_STATE_FAIL: return KNOT_STATE_FAIL;
-		case KNOT_STATE_DONE: return KNOT_STATE_PRODUCE;
+		case KR_STATE_FAIL: return KR_STATE_FAIL;
+		case KR_STATE_DONE: return KR_STATE_PRODUCE;
 		default: break;
 		}
-		DEBUG_MSG(qry, "=> resuming yielded answer\n");
+		VERBOSE_MSG(qry, "=> resuming yielded answer\n");
 		struct kr_layer_pickle *pickle = qry->deferred;
-		request->state = KNOT_STATE_YIELD;
+		request->state = KR_STATE_YIELD;
+		set_yield(&request->answ_selected, qry->uid, false);
+		set_yield(&request->auth_selected, qry->uid, false);
 		RESUME_LAYERS(layer_id(request, pickle->api), request, qry, consume, pickle->pkt);
 		qry->deferred = pickle->next;
 	} else {
@@ -748,45 +971,48 @@ int kr_resolve_produce(struct kr_request *request, struct sockaddr **dst, int *t
 		 * this is normally not required, and incurrs another cache lookups for cached answer. */
 		if (qry->flags & QUERY_ALWAYS_CUT) {
 			switch(zone_cut_check(request, qry, packet)) {
-			case KNOT_STATE_FAIL: return KNOT_STATE_FAIL;
-			case KNOT_STATE_DONE: return KNOT_STATE_PRODUCE;
+			case KR_STATE_FAIL: return KR_STATE_FAIL;
+			case KR_STATE_DONE: return KR_STATE_PRODUCE;
 			default: break;
 			}
 		}
 		/* Resolve current query and produce dependent or finish */
-		request->state = KNOT_STATE_PRODUCE;
+		request->state = KR_STATE_PRODUCE;
 		ITERATE_LAYERS(request, qry, produce, packet);
-		if (request->state != KNOT_STATE_FAIL && knot_wire_get_qr(packet->wire)) {
+		if (request->state != KR_STATE_FAIL && knot_wire_get_qr(packet->wire)) {
 			/* Produced an answer, consume it. */
 			qry->secret = 0;
-			request->state = KNOT_STATE_CONSUME;
+			request->state = KR_STATE_CONSUME;
 			ITERATE_LAYERS(request, qry, consume, packet);
 		}
 	}
 	switch(request->state) {
-	case KNOT_STATE_FAIL: return request->state;
-	case KNOT_STATE_CONSUME: break;
-	case KNOT_STATE_DONE:
+	case KR_STATE_FAIL: return request->state;
+	case KR_STATE_CONSUME: break;
+	case KR_STATE_DONE:
 	default: /* Current query is done */
-		if (qry->flags & QUERY_RESOLVED && request->state != KNOT_STATE_YIELD) {
+		if (qry->flags & QUERY_RESOLVED && request->state != KR_STATE_YIELD) {
 			kr_rplan_pop(rplan, qry);
 		}
 		ITERATE_LAYERS(request, qry, reset);
-		return kr_rplan_empty(rplan) ? KNOT_STATE_DONE : KNOT_STATE_PRODUCE;
+		return kr_rplan_empty(rplan) ? KR_STATE_DONE : KR_STATE_PRODUCE;
 	}
+	
 
 	/* This query has RD=0 or is ANY, stop here. */
 	if (qry->stype == KNOT_RRTYPE_ANY || !knot_wire_get_rd(request->answer->wire)) {
-		DEBUG_MSG(qry, "=> qtype is ANY or RD=0, bail out\n");
-		return KNOT_STATE_FAIL;
+		VERBOSE_MSG(qry, "=> qtype is ANY or RD=0, bail out\n");
+		return KR_STATE_FAIL;
 	}
 
 	/* Update zone cut, spawn new subrequests. */
-	int state = zone_cut_check(request, qry, packet);
-	switch(state) {
-	case KNOT_STATE_FAIL: return KNOT_STATE_FAIL;
-	case KNOT_STATE_DONE: return KNOT_STATE_PRODUCE;
-	default: break;
+	if (!(qry->flags & QUERY_STUB)) {
+		int state = zone_cut_check(request, qry, packet);
+		switch(state) {
+		case KR_STATE_FAIL: return KR_STATE_FAIL;
+		case KR_STATE_DONE: return KR_STATE_PRODUCE;
+		default: break;
+		}
 	}
 
 ns_election:
@@ -796,13 +1022,14 @@ ns_election:
 	 * elect best address only, otherwise elect a completely new NS.
 	 */
 	if(++ns_election_iter >= KR_ITER_LIMIT) {
-		DEBUG_MSG(qry, "=> couldn't converge NS selection, bail out\n");
-		return KNOT_STATE_FAIL;
+		VERBOSE_MSG(qry, "=> couldn't converge NS selection, bail out\n");
+		return KR_STATE_FAIL;
 	}
 
+	const bool retry = (qry->flags & (QUERY_TCP|QUERY_STUB|QUERY_BADCOOKIE_AGAIN));
 	if (qry->flags & (QUERY_AWAIT_IPV4|QUERY_AWAIT_IPV6)) {
 		kr_nsrep_elect_addr(qry, request->ctx);
-	} else if (!qry->ns.name || !(qry->flags & (QUERY_TCP|QUERY_STUB|QUERY_BADCOOKIE_AGAIN))) { /* Keep NS when requerying/stub/badcookie. */
+	} else if (!qry->ns.name || !retry) { /* Keep NS when requerying/stub/badcookie. */
 		/* Root DNSKEY must be fetched from the hints to avoid chicken and egg problem. */
 		if (qry->sname[0] == '\0' && qry->stype == KNOT_RRTYPE_DNSKEY) {
 			kr_zonecut_set_sbelt(request->ctx, &qry->zone_cut);
@@ -810,10 +1037,10 @@ ns_election:
 		}
 		kr_nsrep_elect(qry, request->ctx);
 		if (qry->ns.score > KR_NS_MAX_SCORE) {
-			DEBUG_MSG(qry, "=> no valid NS left\n");
+			VERBOSE_MSG(qry, "=> no valid NS left\n");
 			ITERATE_LAYERS(request, qry, reset);
 			kr_rplan_pop(rplan, qry);
-			return KNOT_STATE_PRODUCE;
+			return KR_STATE_PRODUCE;
 		}
 	}
 
@@ -825,9 +1052,14 @@ ns_election:
 			goto ns_election; /* Must try different NS */
 		}
 		ITERATE_LAYERS(request, qry, reset);
-		return KNOT_STATE_PRODUCE;
+		return KR_STATE_PRODUCE;
 	}
 
+	/* Randomize query case (if not in safemode) */
+	qry->secret = (qry->flags & QUERY_SAFEMODE) ? 0 : kr_rand_uint(UINT32_MAX);
+	knot_dname_t *qname_raw = (knot_dname_t *)knot_pkt_qname(packet);
+	randomized_qname_case(qname_raw, qry->secret);
+
 	/*
 	 * Additional query is going to be finalised when calling
 	 * kr_resolve_checkout().
@@ -907,7 +1139,7 @@ int kr_resolve_checkout(struct kr_request *request, struct sockaddr *src,
 		return kr_error(EINVAL);
 	}
 
-	WITH_DEBUG {
+	WITH_VERBOSE {
 	char qname_str[KNOT_DNAME_MAXLEN], zonecut_str[KNOT_DNAME_MAXLEN], ns_str[INET6_ADDRSTRLEN], type_str[16];
 	knot_dname_to_str(qname_str, knot_pkt_qname(packet), sizeof(qname_str));
 	knot_dname_to_str(zonecut_str, qry->zone_cut.name, sizeof(zonecut_str));
@@ -917,27 +1149,29 @@ int kr_resolve_checkout(struct kr_request *request, struct sockaddr *src,
 		if (addr->sa_family == AF_UNSPEC) {
 			break;
 		}
+		if (!kr_inaddr_equal(dst, addr)) {
+			continue;
+		}
 		inet_ntop(addr->sa_family, kr_nsrep_inaddr(qry->ns.addr[i]), ns_str, sizeof(ns_str));
-		DEBUG_MSG(qry, "%s: '%s' score: %u zone cut: '%s' m12n: '%s' type: '%s'\n",
-		          i == 0 ? "=> querying" : "   optional",
-		          ns_str, qry->ns.score, zonecut_str, qname_str, type_str);
-	}
-	}
+		VERBOSE_MSG(qry, "=> querying: '%s' score: %u zone cut: '%s' m12n: '%s' type: '%s' proto: '%s'\n",
+			ns_str, qry->ns.score, zonecut_str, qname_str, type_str, (qry->flags & QUERY_TCP) ? "tcp" : "udp");
+		break;
+	}}
 
 	return kr_ok();
 }
 
 int kr_resolve_finish(struct kr_request *request, int state)
 {
-#ifndef NDEBUG
+#ifndef NOVERBOSELOG
 	struct kr_rplan *rplan = &request->rplan;
 #endif
 	/* Finalize answer */
 	if (answer_finalize(request, state) != 0) {
-		state = KNOT_STATE_FAIL;
+		state = KR_STATE_FAIL;
 	}
 	/* Error during procesing, internal failure */
-	if (state != KNOT_STATE_DONE) {
+	if (state != KR_STATE_DONE) {
 		knot_pkt_t *answer = request->answer;
 		if (knot_wire_get_rcode(answer->wire) == KNOT_RCODE_NOERROR) {
 			knot_wire_set_rcode(answer->wire, KNOT_RCODE_SERVFAIL);
@@ -946,9 +1180,9 @@ int kr_resolve_finish(struct kr_request *request, int state)
 
 	request->state = state;
 	ITERATE_LAYERS(request, NULL, finish);
-	DEBUG_MSG(NULL, "finished: %d, queries: %zu, mempool: %zu B\n",
+	VERBOSE_MSG(NULL, "finished: %d, queries: %zu, mempool: %zu B\n",
 	          request->state, rplan->resolved.len, (size_t) mp_total_size(request->pool.ctx));
-	return KNOT_STATE_DONE;
+	return KR_STATE_DONE;
 }
 
 struct kr_rplan *kr_resolve_plan(struct kr_request *request)
@@ -967,5 +1201,4 @@ knot_mm_t *kr_resolve_pool(struct kr_request *request)
 	return NULL;
 }
 
-#undef DEBUG_MSG
-
+#undef VERBOSE_MSG
