@@ -1,4 +1,4 @@
-/*  Copyright (C) 2014 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -39,7 +39,8 @@
 	X(NO_CACHE,        1 << 11) /**< Do not use expiring cache for lookup. */ \
 	X(EXPIRING,        1 << 12) /**< Query response is cached, but expiring. */ \
 	X(ALLOW_LOCAL,     1 << 13) /**< Allow queries to local or private address ranges. */ \
-	X(DNSSEC_WANT,     1 << 14) /**< Want DNSSEC secured answer. */ \
+	X(DNSSEC_WANT,     1 << 14) /**< Want DNSSEC secured answer; exception: +cd, \
+				     *   i.e. knot_wire_set_cd(request->answer->wire). */ \
 	X(DNSSEC_BOGUS,    1 << 15) /**< Query response is DNSSEC bogus. */ \
 	X(DNSSEC_INSECURE, 1 << 16) /**< Query response is DNSSEC insecure. */ \
 	X(STUB,            1 << 17) /**< Stub resolution, accept received answer as solved. */ \
@@ -47,7 +48,11 @@
 	X(DNSSEC_WEXPAND,  1 << 19) /**< Query response has wildcard expansion. */ \
 	X(PERMISSIVE,      1 << 20) /**< Permissive resolver mode. */ \
 	X(STRICT,          1 << 21) /**< Strict resolver mode. */ \
-	X(BADCOOKIE_AGAIN, 1 << 22) /**< Query again because bad cookie returned. */
+	X(BADCOOKIE_AGAIN, 1 << 22) /**< Query again because bad cookie returned. */ \
+	X(CNAME,           1 << 23) /**< Query response contains CNAME in answer section. */ \
+	X(REORDER_RR,      1 << 24) /**< Reorder cached RRs. */ \
+	X(TRACE,	   1 << 25) /**< Log answer with kr_verbose_log(), unless -DNDEBUG. */
+			/* 1 << 31       Used by ../modules/dns64/dns64.lua */
 
 /** Query flags */
 enum kr_query_flag {
@@ -72,10 +77,12 @@ struct kr_query {
 	uint32_t flags;
 	uint32_t secret;
 	uint16_t fails;
+	uint16_t reorder; /**< Seed to reorder (cached) RRs in answer or zero. */
 	struct timeval timestamp;
 	struct kr_zonecut zone_cut;
 	struct kr_nsrep ns;
 	struct kr_layer_pickle *deferred;
+	uint32_t uid; /**< Query iteration number, unique within the kr_rplan. */
 };
 
 /** @cond internal Array of queries. */
@@ -93,7 +100,8 @@ struct kr_rplan {
 	kr_qarray_t pending;        /**< List of pending queries. */
 	kr_qarray_t resolved;       /**< List of resolved queries. */
 	struct kr_request *request; /**< Parent resolution request. */
-	knot_mm_t *pool;             /**< Temporary memory pool. */
+	knot_mm_t *pool;            /**< Temporary memory pool. */
+	uint32_t next_uid;          /**< Next value for kr_query::uid (incremental). */
 };
 
 /**
