@@ -1,4 +1,4 @@
-/*  Copyright (C) 2014 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -34,39 +34,49 @@ struct hint_info {
 	const uint8_t *addr;
 };
 
-/* Initialize with SBELT name servers. */
 #define U8(x) (const uint8_t *)(x)
-#define I4 sizeof(struct in_addr)
-#define I6 sizeof(struct in6_addr)
-#define HINT_COUNT 24
-static const struct hint_info SBELT[HINT_COUNT] = {
-        { U8("\x01""j""\x0c""root-servers""\x03""net"), I4, U8("\xc0:\x80\x1e")    }, /* 192.58.128.30 */
-        { U8("\x01""k""\x0c""root-servers""\x03""net"), I4, U8("\xc1\x00\x0e\x81") }, /* 193.0.14.129 */
-        { U8("\x01""d""\x0c""root-servers""\x03""net"), I4, U8("\xc7\x07[\r")      }, /* 199.7.91.13 */
-        { U8("\x01""e""\x0c""root-servers""\x03""net"), I4, U8("\xc0\xcb\xe6\n")   }, /* 192.203.230.10 */
-        { U8("\x01""f""\x0c""root-servers""\x03""net"), I4, U8("\xc0\x05\x05\xf1") }, /* 192.5.5.241 */
-        { U8("\x01""g""\x0c""root-servers""\x03""net"), I4, U8("\xc0p$\x04")       }, /* 192.112.36.4 */
-        { U8("\x01""h""\x0c""root-servers""\x03""net"), I4, U8("\xc6\x61\xbe\x35") }, /* 198.97.190.53 */
-        { U8("\x01""i""\x0c""root-servers""\x03""net"), I4, U8("\xc0$\x94\x11")    }, /* 192.36.148.17 */
-        { U8("\x01""l""\x0c""root-servers""\x03""net"), I4, U8("\xc7\x07S*")       }, /* 199.7.83.42 */
-        { U8("\x01""m""\x0c""root-servers""\x03""net"), I4, U8("\xca\x0c\x1b!")    }, /* 202.12.27.33 */
-        { U8("\x01""b""\x0c""root-servers""\x03""net"), I4, U8("\xc0\xe4O\xc9")    }, /* 192.228.79.201 */
-        { U8("\x01""c""\x0c""root-servers""\x03""net"), I4, U8("\xc6)\x00\x04")    }, /* 192.33.4.12 */
-        { U8("\x01""a""\x0c""root-servers""\x03""net"), I4, U8("\xc6)\x00\x04")    }, /* 198.41.0.4 */
-        { U8("\x01""j""\x0c""root-servers""\x03""net"), I6, U8(" \x01\x05\x03\x0c'\x00\x00\x00\x00\x00\x00\x00\x02\x00\x03")    },
-        { U8("\x01""k""\x0c""root-servers""\x03""net"), I6, U8(" \x01\x07\xfd\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01") },
-        { U8("\x01""d""\x0c""root-servers""\x03""net"), I6, U8(" \x01\x05\x00\x00-\x00\x00\x00\x00\x00\x00\x00\x00\x00\r")      },
-        { U8("\x01""f""\x0c""root-servers""\x03""net"), I6, U8(" \x01\x05\x00\x00/\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0f")    },
-        { U8("\x01""h""\x0c""root-servers""\x03""net"), I6, U8(" \x01\x05\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00S")    },
-        { U8("\x01""i""\x0c""root-servers""\x03""net"), I6, U8(" \x01\x07\xfe\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00S")    },
-        { U8("\x01""l""\x0c""root-servers""\x03""net"), I6, U8(" \x01\x05\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00B")    },
-        { U8("\x01""m""\x0c""root-servers""\x03""net"), I6, U8(" \x01\r\xc3\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x005")      },
-        { U8("\x01""b""\x0c""root-servers""\x03""net"), I6, U8(" \x01\x05\x00\x00\x84\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0b") },
-        { U8("\x01""c""\x0c""root-servers""\x03""net"), I6, U8(" \x01\x05\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0c") },
-        { U8("\x01""a""\x0c""root-servers""\x03""net"), I6, U8(" \x01\x05\x03\xba>\x00\x00\x00\x00\x00\x00\x00\x02\x000")       },
+
+/* Initialize with SBELT name servers. */
+#include "lib/root-hints.inc"
+#define HINT_LEN_A sizeof(struct in_addr)
+#define HINT_LEN_AAAA sizeof(struct in6_addr)
+#define SBELT_LEN (sizeof(SBELT) / sizeof(SBELT[0]))
+static const struct hint_info SBELT[] = {
+	#define HINT(name,type) { \
+		U8("\x01" #name "\x0c""root-servers""\x03""net"), \
+		HINT_LEN_##type, \
+		U8(HINT_##name##_##type), \
+      	}
+	HINT(j, A),
+	HINT(k, A),
+	HINT(d, A),
+	HINT(e, A),
+	HINT(f, A),
+	HINT(g, A),
+	HINT(h, A),
+	HINT(i, A),
+	HINT(l, A),
+	HINT(m, A),
+	HINT(b, A),
+	HINT(c, A),
+	HINT(a, A),
+	HINT(a, AAAA),
+	HINT(b, AAAA),
+	HINT(c, AAAA),
+	HINT(d, AAAA),
+	HINT(e, AAAA),
+	HINT(f, AAAA),
+	HINT(g, AAAA),
+	HINT(h, AAAA),
+	HINT(i, AAAA),
+	HINT(j, AAAA),
+	HINT(k, AAAA),
+	HINT(l, AAAA),
+	HINT(m, AAAA),
+	#undef HINT
 };
-#undef I4
-#undef I6
+#undef HINT_LEN_A
+#undef HINT_LEN_AAAA
 
 
 static void update_cut_name(struct kr_zonecut *cut, const knot_dname_t *name)
@@ -257,6 +267,21 @@ int kr_zonecut_del(struct kr_zonecut *cut, const knot_dname_t *ns, const knot_rd
 	return ret;
 }
 
+int kr_zonecut_del_all(struct kr_zonecut *cut, const knot_dname_t *ns)
+{
+	if (!cut || !ns) {
+		return kr_error(EINVAL);
+	}
+
+	/* Find the address list; then free and remove it. */
+	pack_t *pack = kr_zonecut_find(cut, ns);
+	if (pack == NULL) {
+		return kr_error(ENOENT);
+	}
+	free_addr_set((const char *)ns, pack, cut->pool);
+	return map_del(&cut->nsset, (const char *)ns);
+}
+
 pack_t *kr_zonecut_find(struct kr_zonecut *cut, const knot_dname_t *ns)
 {
 	if (!cut || !ns) {
@@ -286,7 +311,7 @@ int kr_zonecut_set_sbelt(struct kr_context *ctx, struct kr_zonecut *cut)
 		ret = kr_zonecut_copy(cut, &ctx->root_hints);
 	} else {
 		/* Copy compiled-in root hints */
-		for (unsigned i = 0; i < HINT_COUNT; ++i) {
+		for (unsigned i = 0; i < SBELT_LEN; ++i) {
 			const struct hint_info *hint = &SBELT[i];
 			knot_rdata_init(rdata_arr, hint->len, hint->addr, 0);
 			ret = kr_zonecut_add(cut, hint->name, rdata_arr);
@@ -330,7 +355,7 @@ static int fetch_ns(struct kr_context *ctx, struct kr_zonecut *cut, const knot_d
 
 	/* Materialize as we'll going to do more cache lookups. */
 	knot_rrset_t rr_copy;
-	ret = kr_cache_materialize(&rr_copy, &cached_rr, drift, cut->pool);
+	ret = kr_cache_materialize(&rr_copy, &cached_rr, drift, 0, cut->pool);
 	if (ret != 0) {
 		return ret;
 	}
@@ -341,7 +366,8 @@ static int fetch_ns(struct kr_context *ctx, struct kr_zonecut *cut, const knot_d
 		const knot_dname_t *ns_name = knot_ns_name(&rr_copy.rrs, i);
 		kr_zonecut_add(cut, ns_name, NULL);
 		/* Fetch NS reputation and decide whether to prefetch A/AAAA records. */
-		unsigned *cached = lru_get(ctx->cache_rep, (const char *)ns_name, knot_dname_size(ns_name));
+		unsigned *cached = lru_get_try(ctx->cache_rep,
+				(const char *)ns_name, knot_dname_size(ns_name));
 		unsigned reputation = (cached) ? *cached : 0;
 		if (!(reputation & KR_NS_NOIP4) && !(ctx->options & QUERY_NO_IPV4)) {
 			fetch_addr(cut, &ctx->cache, ns_name, KNOT_RRTYPE_A, timestamp);
@@ -380,7 +406,7 @@ static int fetch_rrset(knot_rrset_t **rr, struct kr_cache *cache,
 		return kr_error(ENOMEM);
 	}
 
-	ret = kr_cache_materialize(*rr, &cached_rr, drift, pool);
+	ret = kr_cache_materialize(*rr, &cached_rr, drift, 0, pool);
 	if (ret != 0) {
 		knot_rrset_free(rr, pool);
 		return ret;
